# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""Various conic constraint implementations."""

from pyomo.common.modeling import NOTSET
from pyomo.core.expr.numvalue import is_numeric_data
from pyomo.core.expr import value, exp
from pyomo.core.expr.expr_common import _type_check_exception_arg
from pyomo.core.kernel.block import block
from pyomo.core.kernel.variable import IVariable, variable, variable_tuple
from pyomo.core.kernel.constraint import (
    IConstraint,
    linear_constraint,
    constraint,
    constraint_tuple,
)


def _build_linking_constraints(v, v_aux):
    assert len(v) == len(v_aux)
    c_aux = []
    for vi, vi_aux in zip(v, v_aux):
        assert vi_aux.ctype is IVariable
        if vi is None:
            continue
        elif is_numeric_data(vi):
            c_aux.append(
                linear_constraint(variables=(vi_aux,), coefficients=(1,), rhs=vi)
            )
        elif isinstance(vi, IVariable):
            c_aux.append(
                linear_constraint(variables=(vi_aux, vi), coefficients=(1, -1), rhs=0)
            )
        else:
            c_aux.append(constraint(body=vi_aux - vi, rhs=0))
    return constraint_tuple(c_aux)


class _ConicBase(IConstraint):
    """Base class for a few conic constraints that
    implements some shared functionality. Derived classes
    are expected to declare any necessary slots."""

    _ctype = IConstraint
    _linear_canonical_form = False
    __slots__ = ()

    def __init__(self):
        self._parent = None
        self._storage_key = None
        self._active = True
        # the body expression is only built if necessary
        # (i.e., when someone asks for it via the body
        # property method)
        self._body = None

    @classmethod
    def as_domain(cls, *args, **kwds):
        """Builds a conic domain"""
        raise NotImplementedError  # pragma:nocover

    def _body_function(self, *args):
        """A function that defines the body expression"""
        raise NotImplementedError  # pragma:nocover

    def _body_function_variables(self, values=False):
        """Returns variables in the order they should be
        passed to the body function. If values is True, then
        return the current value of each variable in place
        of the variables themselves."""
        raise NotImplementedError  # pragma:nocover

    def check_convexity_conditions(self, relax=False):
        """Returns True if all convexity conditions for the
        conic constraint are satisfied. If relax is True,
        then variable domains are ignored and it is assumed
        that all variables are continuous."""
        raise NotImplementedError  # pragma:nocover

    #
    # Define the IConstraint abstract methods
    #

    @property
    def body(self):
        """The body of the constraint"""
        if self._body is None:
            self._body = self._body_function(
                *self._body_function_variables(values=False)
            )
        return self._body

    @property
    def lower(self):
        """The expression for the lower bound of the constraint"""
        return None

    @property
    def upper(self):
        """The expression for the upper bound of the constraint"""
        return 0.0

    @property
    def lb(self):
        """The value of the lower bound of the constraint"""
        return None

    @property
    def ub(self):
        """The value of the upper bound of the constraint"""
        return 0.0

    @property
    def rhs(self):
        """The right-hand side of the constraint"""
        raise ValueError(
            "The rhs property can not be read because this "
            "is not an equality constraint"
        )

    @property
    def equality(self):
        return False

    #
    # Override a the default __call__ method on IConstraint
    # to avoid building the body expression, if possible
    #

    def __call__(self, exception=NOTSET):
        exception = _type_check_exception_arg(self, exception)
        try:
            # we wrap the result with value(...) as the
            # alpha term used by some of the constraints
            # may be a parameter
            return value(
                self._body_function(*self._body_function_variables(values=True))
            )
        except (ValueError, TypeError):
            if exception:
                raise ValueError("one or more terms could not be evaluated")
            return None


class quadratic(_ConicBase):
    """A quadratic conic constraint of the form:

    .. math::

        x[0]^2 + ... + x[n-1]^2 <= r^2,

    which is recognized as convex for r >= 0.

    Parameters
    ----------
    r : :class:`variable`
        A variable.
    x : list[:class:`variable`]
        An iterable of variables.
    """

    __slots__ = (
        "_parent",
        "_storage_key",
        "_active",
        "_body",
        "_r",
        "_x",
        "__weakref__",
    )

    def __init__(self, r, x):
        super(quadratic, self).__init__()
        self._r = r
        self._x = tuple(x)
        assert isinstance(self._r, IVariable)
        assert all(isinstance(xi, IVariable) for xi in self._x)

    @classmethod
    def as_domain(cls, r, x):
        """Builds a conic domain. Input arguments take the
        same form as those of the conic constraint, but in
        place of each variable, one can optionally supply a
        constant, linear expression, or None.

        Returns
        -------
        block
            A block object with the core conic constraint
            (block.q) expressed using auxiliary variables
            (block.r, block.x) linked to the input arguments
            through auxiliary constraints (block.c).
        """
        b = block()
        b.r = variable(lb=0)
        b.x = variable_tuple([variable() for i in range(len(x))])
        b.c = _build_linking_constraints([r] + list(x), [b.r] + list(b.x))
        b.q = cls(r=b.r, x=b.x)
        return b

    @property
    def r(self):
        return self._r

    @property
    def x(self):
        return self._x

    #
    # Define the _ConicBase abstract methods
    #

    def _body_function(self, r, x):
        """A function that defines the body expression"""
        return sum(xi**2 for xi in x) - r**2

    def _body_function_variables(self, values=False):
        """Returns variables in the order they should be
        passed to the body function. If values is True, then
        return the current value of each variable in place
        of the variables themselves."""
        if not values:
            return self.r, self.x
        else:
            return self.r.value, tuple(xi.value for xi in self.x)

    def check_convexity_conditions(self, relax=False):
        """Returns True if all convexity conditions for the
        conic constraint are satisfied. If relax is True,
        then variable domains are ignored and it is assumed
        that all variables are continuous."""
        return (
            relax
            or (self.r.is_continuous() and all(xi.is_continuous() for xi in self.x))
        ) and (self.r.has_lb() and value(self.r.lb) >= 0)


class rotated_quadratic(_ConicBase):
    """A rotated quadratic conic constraint of the form:

    .. math::

        x[0]^2 + ... + x[n-1]^2 <= 2*r1*r2,

    which is recognized as convex for r1,r2 >= 0.

    Parameters
    ----------
    r1 : :class:`variable`
        A variable.
    r2 : :class:`variable`
        A variable.
    x : list[:class:`variable`]
        An iterable of variables.
    """

    __slots__ = (
        "_parent",
        "_storage_key",
        "_active",
        "_body",
        "_r1",
        "_r2",
        "_x",
        "__weakref__",
    )

    def __init__(self, r1, r2, x):
        super(rotated_quadratic, self).__init__()
        self._r1 = r1
        self._r2 = r2
        self._x = tuple(x)
        assert isinstance(self._r1, IVariable)
        assert isinstance(self._r2, IVariable)
        assert all(isinstance(xi, IVariable) for xi in self._x)

    @classmethod
    def as_domain(cls, r1, r2, x):
        """Builds a conic domain. Input arguments take the
        same form as those of the conic constraint, but in
        place of each variable, one can optionally supply a
        constant, linear expression, or None.

        Returns
        -------
        block
            A block object with the core conic constraint
            (block.q) expressed using auxiliary variables
            (block.r1, block.r2, block.x) linked to the
            input arguments through auxiliary constraints
            (block.c).
        """
        b = block()
        b.r1 = variable(lb=0)
        b.r2 = variable(lb=0)
        b.x = variable_tuple([variable() for i in range(len(x))])
        b.c = _build_linking_constraints([r1, r2] + list(x), [b.r1, b.r2] + list(b.x))
        b.q = cls(r1=b.r1, r2=b.r2, x=b.x)
        return b

    @property
    def r1(self):
        return self._r1

    @property
    def r2(self):
        return self._r2

    @property
    def x(self):
        return self._x

    #
    # Define the _ConicBase abstract methods
    #

    def _body_function(self, r1, r2, x):
        """A function that defines the body expression"""
        return sum(xi**2 for xi in x) - 2 * r1 * r2

    def _body_function_variables(self, values=False):
        """Returns variables in the order they should be
        passed to the body function. If values is True, then
        return the current value of each variable in place
        of the variables themselves."""
        if not values:
            return self.r1, self.r2, self.x
        else:
            return self.r1.value, self.r2.value, tuple(xi.value for xi in self.x)

    def check_convexity_conditions(self, relax=False):
        """Returns True if all convexity conditions for the
        conic constraint are satisfied. If relax is True,
        then variable domains are ignored and it is assumed
        that all variables are continuous."""
        return (
            (
                relax
                or (
                    self.r1.is_continuous()
                    and self.r2.is_continuous()
                    and all(xi.is_continuous() for xi in self.x)
                )
            )
            and (self.r1.has_lb() and value(self.r1.lb) >= 0)
            and (self.r2.has_lb() and value(self.r2.lb) >= 0)
        )


class primal_exponential(_ConicBase):
    """A primal exponential conic constraint of the form:

    .. math::

        x1*exp(x2/x1) <= r,

    which is recognized as convex for x1,r >= 0.

    Parameters
    ----------
    r : :class:`variable`
        A variable.
    x1 : :class:`variable`
        A variable.
    x2 : :class:`variable`
        A variable.
    """

    __slots__ = (
        "_parent",
        "_storage_key",
        "_active",
        "_body",
        "_r",
        "_x1",
        "_x2",
        "__weakref__",
    )

    def __init__(self, r, x1, x2):
        super(primal_exponential, self).__init__()
        self._r = r
        self._x1 = x1
        self._x2 = x2
        assert isinstance(self._r, IVariable)
        assert isinstance(self._x1, IVariable)
        assert isinstance(self._x2, IVariable)

    @classmethod
    def as_domain(cls, r, x1, x2):
        """Builds a conic domain. Input arguments take the
        same form as those of the conic constraint, but in
        place of each variable, one can optionally supply a
        constant, linear expression, or None.

        Returns
        -------
        block
            A block object with the core conic constraint
            (block.q) expressed using auxiliary variables
            (block.r, block.x1, block.x2) linked to the
            input arguments through auxiliary constraints
            (block.c).
        """
        b = block()
        b.r = variable(lb=0)
        b.x1 = variable(lb=0)
        b.x2 = variable()
        b.c = _build_linking_constraints([r, x1, x2], [b.r, b.x1, b.x2])
        b.q = cls(r=b.r, x1=b.x1, x2=b.x2)
        return b

    @property
    def r(self):
        return self._r

    @property
    def x1(self):
        return self._x1

    @property
    def x2(self):
        return self._x2

    #
    # Define the _ConicBase abstract methods
    #

    def _body_function(self, r, x1, x2):
        """A function that defines the body expression"""
        return x1 * exp(x2 / x1) - r

    def _body_function_variables(self, values=False):
        """Returns variables in the order they should be
        passed to the body function. If values is True, then
        return the current value of each variable in place
        of the variables themselves."""
        if not values:
            return self.r, self.x1, self.x2
        else:
            return self.r.value, self.x1.value, self.x2.value

    def check_convexity_conditions(self, relax=False):
        """Returns True if all convexity conditions for the
        conic constraint are satisfied. If relax is True,
        then variable domains are ignored and it is assumed
        that all variables are continuous."""
        return (
            (
                relax
                or (
                    self.x1.is_continuous()
                    and self.x2.is_continuous()
                    and self.r.is_continuous()
                )
            )
            and (self.x1.has_lb() and value(self.x1.lb) >= 0)
            and (self.r.has_lb() and value(self.r.lb) >= 0)
        )


class primal_power(_ConicBase):
    """A primal power conic constraint of the form:

    .. math::

       sqrt(x[0]^2 + ... + x[n-1]^2) <= (r1^alpha)*(r2^(1-alpha))

    which is recognized as convex for r1,r2 >= 0
    and 0 < alpha < 1.

    Parameters
    ----------
    r1 : :class:`variable`
        A variable.
    r2 : :class:`variable`
        A variable.
    x : list[:class:`variable`]
        An iterable of variables.
    alpha : float, :class:`parameter`, etc.
        A constant term.
    """

    __slots__ = (
        "_parent",
        "_storage_key",
        "_active",
        "_body",
        "_r1",
        "_r2",
        "_x",
        "_alpha",
        "__weakref__",
    )

    def __init__(self, r1, r2, x, alpha):
        super(primal_power, self).__init__()
        self._r1 = r1
        self._r2 = r2
        self._x = tuple(x)
        self._alpha = alpha
        assert isinstance(self._r1, IVariable)
        assert isinstance(self._r2, IVariable)
        assert all(isinstance(xi, IVariable) for xi in self._x)
        if not is_numeric_data(self._alpha):
            raise TypeError(
                "The type of the alpha parameter of a conic "
                "constraint is restricted numeric data or "
                "objects that store numeric data."
            )

    @classmethod
    def as_domain(cls, r1, r2, x, alpha):
        """Builds a conic domain. Input arguments take the
        same form as those of the conic constraint, but in
        place of each variable, one can optionally supply a
        constant, linear expression, or None.

        Returns
        -------
        block
            A block object with the core conic constraint
            (block.q) expressed using auxiliary variables
            (block.r1, block.r2, block.x) linked to the
            input arguments through auxiliary constraints
            (block.c).
        """
        b = block()
        b.r1 = variable(lb=0)
        b.r2 = variable(lb=0)
        b.x = variable_tuple([variable() for i in range(len(x))])
        b.c = _build_linking_constraints([r1, r2] + list(x), [b.r1, b.r2] + list(b.x))
        b.q = cls(r1=b.r1, r2=b.r2, x=b.x, alpha=alpha)
        return b

    @property
    def r1(self):
        return self._r1

    @property
    def r2(self):
        return self._r2

    @property
    def x(self):
        return self._x

    @property
    def alpha(self):
        return self._alpha

    #
    # Define the _ConicBase abstract methods
    #

    def _body_function(self, r1, r2, x):
        """A function that defines the body expression"""
        alpha = self.alpha
        return (sum(xi**2 for xi in x) ** 0.5) - (r1**alpha) * (r2 ** (1 - alpha))

    def _body_function_variables(self, values=False):
        """Returns variables in the order they should be
        passed to the body function. If values is True, then
        return the current value of each variable in place
        of the variables themselves."""
        if not values:
            return self.r1, self.r2, self.x
        else:
            return self.r1.value, self.r2.value, tuple(xi.value for xi in self.x)

    def check_convexity_conditions(self, relax=False):
        """Returns True if all convexity conditions for the
        conic constraint are satisfied. If relax is True,
        then variable domains are ignored and it is assumed
        that all variables are continuous."""
        alpha = value(self.alpha, exception=False)
        return (
            (
                relax
                or (
                    self.r1.is_continuous()
                    and self.r2.is_continuous()
                    and all(xi.is_continuous() for xi in self.x)
                )
            )
            and (self.r1.has_lb() and value(self.r1.lb) >= 0)
            and (self.r2.has_lb() and value(self.r2.lb) >= 0)
            and ((alpha is not None) and (0 < alpha < 1))
        )


class primal_geomean(_ConicBase):
    """A primal geometric mean conic constraint of the form:

    .. math::

        (r[0]*...*r[n-2])^(1/(n-1)) >= |x[n-1]|

    Parameters
    ----------
    r : :class:`variable`
        An iterable of variables.
    x : :class:`variable`
        A scalar variable.

    """

    __slots__ = (
        "_parent",
        "_storage_key",
        "_active",
        "_body",
        "_r",
        "_x",
        "__weakref__",
    )

    def __init__(self, r, x):
        super(primal_geomean, self).__init__()
        self._r = tuple(r)
        self._x = x
        assert isinstance(self._x, IVariable)
        assert all(isinstance(ri, IVariable) for ri in self._r)

    @classmethod
    def as_domain(cls, r, x):
        """Builds a conic domain. Input arguments take the
        same form as those of the conic constraint, but in
        place of each variable, one can optionally supply a
        constant, linear expression, or None.

        Returns
        -------
        block
            A block object with the core conic constraint
            (block.q) expressed using auxiliary variables
            (block.r, block.x) linked to the input arguments
            through auxiliary constraints (block.c)."""
        b = block()
        b.r = variable_tuple([variable(lb=0) for i in range(len(r))])
        b.x = variable()
        b.c = _build_linking_constraints(list(r) + [x], list(b.r) + [b.x])
        b.q = cls(r=b.r, x=b.x)
        return b

    @property
    def r(self):
        return self._r

    @property
    def x(self):
        return self._x


class dual_exponential(_ConicBase):
    """A dual exponential conic constraint of the form:

    .. math::

        -x2*exp((x1/x2)-1) <= r

    which is recognized as convex for x2 <= 0 and r >= 0.

    Parameters
    ----------
    r : :class:`variable`
        A variable.
    x1 : :class:`variable`
        A variable.
    x2 : :class:`variable`
        A variable.
    """

    __slots__ = (
        "_parent",
        "_storage_key",
        "_active",
        "_body",
        "_r",
        "_x1",
        "_x2",
        "__weakref__",
    )

    def __init__(self, r, x1, x2):
        super(dual_exponential, self).__init__()
        self._r = r
        self._x1 = x1
        self._x2 = x2
        assert isinstance(self._r, IVariable)
        assert isinstance(self._x1, IVariable)
        assert isinstance(self._x2, IVariable)

    @classmethod
    def as_domain(cls, r, x1, x2):
        """Builds a conic domain. Input arguments take the
        same form as those of the conic constraint, but in
        place of each variable, one can optionally supply a
        constant, linear expression, or None.

        Returns
        -------
        block
            A block object with the core conic constraint
            (block.q) expressed using auxiliary variables
            (block.r, block.x1, block.x2) linked to the
            input arguments through auxiliary constraints
            (block.c).
        """
        b = block()
        b.r = variable(lb=0)
        b.x1 = variable()
        b.x2 = variable(ub=0)
        b.c = _build_linking_constraints([r, x1, x2], [b.r, b.x1, b.x2])
        b.q = cls(r=b.r, x1=b.x1, x2=b.x2)
        return b

    @property
    def r(self):
        return self._r

    @property
    def x1(self):
        return self._x1

    @property
    def x2(self):
        return self._x2

    #
    # Define the _ConicBase abstract methods
    #

    def _body_function(self, r, x1, x2):
        """A function that defines the body expression"""
        return -x2 * exp((x1 / x2) - 1) - r

    def _body_function_variables(self, values=False):
        """Returns variables in the order they should be
        passed to the body function. If values is True, then
        return the current value of each variable in place
        of the variables themselves."""
        if not values:
            return self.r, self.x1, self.x2
        else:
            return self.r.value, self.x1.value, self.x2.value

    def check_convexity_conditions(self, relax=False):
        """Returns True if all convexity conditions for the
        conic constraint are satisfied. If relax is True,
        then variable domains are ignored and it is assumed
        that all variables are continuous."""
        return (
            (
                relax
                or (
                    self.x1.is_continuous()
                    and self.x2.is_continuous()
                    and self.r.is_continuous()
                )
            )
            and (self.x2.has_ub() and value(self.x2.ub) <= 0)
            and (self.r.has_lb() and value(self.r.lb) >= 0)
        )


class dual_power(_ConicBase):
    """A dual power conic constraint of the form:

    .. math::

        sqrt(x[0]^2 + ... + x[n-1]^2)
        <=
        ((r1/alpha)^alpha) * ((r2/(1-alpha))^(1-alpha))

    which is recognized as convex for r1,r2 >= 0
    and 0 < alpha < 1.

    Parameters
    ----------
    r1 : :class:`variable`
        A variable.
    r2 : :class:`variable`
        A variable.
    x : list[:class:`variable`]
        An iterable of variables.
    alpha : float, :class:`parameter`, etc.
        A constant term.
    """

    __slots__ = (
        "_parent",
        "_storage_key",
        "_active",
        "_body",
        "_r1",
        "_r2",
        "_x",
        "_alpha",
        "__weakref__",
    )

    def __init__(self, r1, r2, x, alpha):
        super(dual_power, self).__init__()
        self._r1 = r1
        self._r2 = r2
        self._x = tuple(x)
        self._alpha = alpha
        assert isinstance(self._r1, IVariable)
        assert isinstance(self._r2, IVariable)
        assert all(isinstance(xi, IVariable) for xi in self._x)
        if not is_numeric_data(self._alpha):
            raise TypeError(
                "The type of the alpha parameter of a conic "
                "constraint is restricted numeric data or "
                "objects that store numeric data."
            )

    @classmethod
    def as_domain(cls, r1, r2, x, alpha):
        """Builds a conic domain. Input arguments take the
        same form as those of the conic constraint, but in
        place of each variable, one can optionally supply a
        constant, linear expression, or None.

        Returns
        -------
        block
            A block object with the core conic constraint
            (block.q) expressed using auxiliary variables
            (block.r1, block.r2, block.x) linked to the
            input arguments through auxiliary constraints
            (block.c).
        """
        b = block()
        b.r1 = variable(lb=0)
        b.r2 = variable(lb=0)
        b.x = variable_tuple([variable() for i in range(len(x))])
        b.c = _build_linking_constraints([r1, r2] + list(x), [b.r1, b.r2] + list(b.x))
        b.q = cls(r1=b.r1, r2=b.r2, x=b.x, alpha=alpha)
        return b

    @property
    def r1(self):
        return self._r1

    @property
    def r2(self):
        return self._r2

    @property
    def x(self):
        return self._x

    @property
    def alpha(self):
        return self._alpha

    #
    # Define the _ConicBase abstract methods
    #

    def _body_function(self, r1, r2, x):
        """A function that defines the body expression"""
        alpha = self.alpha
        return (sum(xi**2 for xi in x) ** 0.5) - ((r1 / alpha) ** alpha) * (
            (r2 / (1 - alpha)) ** (1 - alpha)
        )

    def _body_function_variables(self, values=False):
        """Returns variables in the order they should be
        passed to the body function. If values is True, then
        return the current value of each variable in place
        of the variables themselves."""
        if not values:
            return self.r1, self.r2, self.x
        else:
            return self.r1.value, self.r2.value, tuple(xi.value for xi in self.x)

    def check_convexity_conditions(self, relax=False):
        """Returns True if all convexity conditions for the
        conic constraint are satisfied. If relax is True,
        then variable domains are ignored and it is assumed
        that all variables are continuous."""
        alpha = value(self.alpha, exception=False)
        return (
            (
                relax
                or (
                    self.r1.is_continuous()
                    and self.r2.is_continuous()
                    and all(xi.is_continuous() for xi in self.x)
                )
            )
            and (self.r1.has_lb() and value(self.r1.lb) >= 0)
            and (self.r2.has_lb() and value(self.r2.lb) >= 0)
            and ((alpha is not None) and (0 < alpha < 1))
        )


class dual_geomean(_ConicBase):
    """A dual geometric mean conic constraint of the form:

    .. math::

        (n-1)*(r[0]*...*r[n-2])^(1/(n-1)) >= |x[n-1]|

    Parameters
    ----------
    r : :class:`variable`
        An iterable of variables.
    x : :class:`variable`
        A scalar variable.

    """

    __slots__ = (
        "_parent",
        "_storage_key",
        "_active",
        "_body",
        "_r",
        "_x",
        "__weakref__",
    )

    def __init__(self, r, x):
        super(dual_geomean, self).__init__()
        self._r = tuple(r)
        self._x = x
        assert isinstance(self._x, IVariable)
        assert all(isinstance(ri, IVariable) for ri in self._r)

    @classmethod
    def as_domain(cls, r, x):
        """Builds a conic domain. Input arguments take the
        same form as those of the conic constraint, but in
        place of each variable, one can optionally supply a
        constant, linear expression, or None.

        Returns
        -------
        block
            A block object with the core conic constraint
            (block.q) expressed using auxiliary variables
            (block.r, block.x) linked to the input arguments
            through auxiliary constraints (block.c)."""
        b = block()
        b.r = variable_tuple([variable(lb=0) for i in range(len(r))])
        b.x = variable()
        b.c = _build_linking_constraints(list(r) + [x], list(b.r) + [b.x])
        b.q = cls(r=b.r, x=b.x)
        return b

    @property
    def r(self):
        return self._r

    @property
    def x(self):
        return self._x


class svec_psdcone(_ConicBase):
    r"""A domain consisting of vectorizations of the lower-triangular
    part of a positive semidefinite matrx, with the non-diagonal
    elements additionally rescaled. In other words, if a vector 'x'
    of length :math:`n = d(d+1)/2` belongs to this cone, then the matrix:

    .. math::
       :nowrap:

       \[\begin{array}{rcclcl}
       sMat(x) = [\;\;
          {[} &      x[1],   &  x[2]/\sqrt{2}, &...,&      x[d]/\sqrt{2} & ], \\
          {[} &x[2]/\sqrt{2},&         x[d+1], &...,&   x[2d-1]/\sqrt{2} & ], \\
              &              &    \vdots       &    &                    &    \\
          {[} &x[d]/\sqrt{2},&x[2d-1]/\sqrt{2},&...,&x[d(d+1)/2]/\sqrt{2}& ]
       \;\;]
       \end{array}\]

    will be restricted to be a positive-semidefinite matrix.

    Parameters
    ----------
    x : :class:`variable`
        An iterable of variables with length :math:`d(d+1)/2`.

    """

    __slots__ = ("_parent", "_storage_key", "_active", "_body", "_x", "__weakref__")

    def __init__(self, x):
        super(svec_psdcone, self).__init__()
        self._x = tuple(x)
        assert all(isinstance(xi, IVariable) for xi in self._x)

    @classmethod
    def as_domain(cls, x):
        """Builds a conic domain. Input arguments take the
        same form as those of the conic constraint, but in
        place of each variable, one can optionally supply a
        constant, linear expression, or None.

        Returns
        -------
        block
            A block object with the core conic constraint
            (block.q) expressed using auxiliary variables
            (block.r, block.x) linked to the input arguments
            through auxiliary constraints (block.c)."""
        b = block()
        b.x = variable_tuple([variable() for i in range(len(x))])
        b.c = _build_linking_constraints(list(x), list(b.x))
        b.q = cls(x=b.x)
        return b

    @property
    def x(self):
        return self._x



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import ctypes
import math
import operator

from collections import deque
from operator import itemgetter

from pyomo.common.deprecation import deprecation_warning
from pyomo.common.errors import DeveloperError, InfeasibleConstraintException, MouseTrap
from pyomo.common.numeric_types import (
    native_complex_types,
    native_numeric_types,
    native_types,
    value,
)


from pyomo.core.base import Expression
from pyomo.core.expr import (
    NegationExpression,
    ProductExpression,
    DivisionExpression,
    PowExpression,
    AbsExpression,
    UnaryFunctionExpression,
    MonomialTermExpression,
    LinearExpression,
    SumExpression,
    EqualityExpression,
    InequalityExpression,
    RangedExpression,
    Expr_ifExpression,
    ExternalFunctionExpression,
)
from pyomo.core.expr.visitor import StreamBasedExpressionVisitor, _EvaluationVisitor
from pyomo.repn.util import (
    BeforeChildDispatcher,
    ExitNodeDispatcher,
    ExprType,
    apply_node_operation,
    complex_number_error,
    nan,
    sum_like_expression_types,
    check_constant,
)

_CONSTANT = ExprType.CONSTANT
_MONOMIAL = ExprType.MONOMIAL
_GENERAL = ExprType.GENERAL

# Feasibility tolerance for trivial (fixed) constraints
TOL = 1e-8


def _create_strict_inequality_map(vars_):
    vars_['strict_inequality_map'] = {
        True: vars_['less_than'],
        False: vars_['less_equal'],
        (True, True): (vars_['less_than'], vars_['less_than']),
        (True, False): (vars_['less_than'], vars_['less_equal']),
        (False, True): (vars_['less_equal'], vars_['less_than']),
        (False, False): (vars_['less_equal'], vars_['less_equal']),
    }


class TextNLDebugTemplate:
    unary = {
        'log': 'o43\t#log\n',
        'log10': 'o42\t#log10\n',
        'sin': 'o41\t#sin\n',
        'cos': 'o46\t#cos\n',
        'tan': 'o38\t#tan\n',
        'sinh': 'o40\t#sinh\n',
        'cosh': 'o45\t#cosh\n',
        'tanh': 'o37\t#tanh\n',
        'asin': 'o51\t#asin\n',
        'acos': 'o53\t#acos\n',
        'atan': 'o49\t#atan\n',
        'exp': 'o44\t#exp\n',
        'sqrt': 'o39\t#sqrt\n',
        'asinh': 'o50\t#asinh\n',
        'acosh': 'o52\t#acosh\n',
        'atanh': 'o47\t#atanh\n',
        'ceil': 'o14\t#ceil\n',
        'floor': 'o13\t#floor\n',
    }

    binary_sum = 'o0\t#+\n'
    product = 'o2\t#*\n'
    division = 'o3\t# /\n'
    pow = 'o5\t#^\n'
    abs = 'o15\t# abs\n'
    negation = 'o16\t#-\n'
    nary_sum = 'o54\t# sumlist\n%d\t# (n)\n'
    exprif = 'o35\t# if\n'
    and_expr = 'o21\t# and\n'
    less_than = 'o22\t# lt\n'
    less_equal = 'o23\t# le\n'
    equality = 'o24\t# eq\n'
    external_fcn = 'f%d %d%s\n'
    # NOTE: to support scaling and substitutions, we do NOT include the
    # 'v' or the EOL here:
    var = '%s'
    const = 'n%s\n'
    string = 'h%d:%s\n'
    monomial = product + const + var.replace('%', '%%')
    multiplier = product + const

    _create_strict_inequality_map(vars())


nl_operators = {
    0: (2, operator.add),
    2: (2, operator.mul),
    3: (2, operator.truediv),
    5: (2, operator.pow),
    15: (1, operator.abs),
    16: (1, operator.neg),
    54: (None, lambda *x: sum(x)),
    35: (3, lambda a, b, c: b if a else c),
    21: (2, operator.and_),
    22: (2, operator.lt),
    23: (2, operator.le),
    24: (2, operator.eq),
    43: (1, math.log),
    42: (1, math.log10),
    41: (1, math.sin),
    46: (1, math.cos),
    38: (1, math.tan),
    40: (1, math.sinh),
    45: (1, math.cosh),
    37: (1, math.tanh),
    51: (1, math.asin),
    53: (1, math.acos),
    49: (1, math.atan),
    44: (1, math.exp),
    39: (1, math.sqrt),
    50: (1, math.asinh),
    52: (1, math.acosh),
    47: (1, math.atanh),
    14: (1, math.ceil),
    13: (1, math.floor),
}


def _strip_template_comments(vars_, base_):
    vars_['unary'] = {
        k: v[: v.find('\t#')] + '\n' if v[-1] == '\n' else ''
        for k, v in base_.unary.items()
    }
    for k, v in base_.__dict__.items():
        if type(v) is str and '\t#' in v:
            v_lines = v.split('\n')
            for i, l in enumerate(v_lines):
                comment_start = l.find('\t#')
                if comment_start >= 0:
                    v_lines[i] = l[:comment_start]
            vars_[k] = '\n'.join(v_lines)


def _inv2str(val):
    return f"{val._str() if hasattr(val, '_str') else val}"


# The "standard" text mode template is the debugging template with the
# comments removed
class TextNLTemplate(TextNLDebugTemplate):
    _strip_template_comments(vars(), TextNLDebugTemplate)
    _create_strict_inequality_map(vars())


class NLFragment:
    """This is a mock "component" for the nl portion of a named Expression.

    It is used internally in the writer when requesting symbolic solver
    labels so that we can generate meaningful names for the nonlinear
    portion of an Expression component.

    """

    __slots__ = ('_repn', '_node')

    def __init__(self, repn, node):
        self._repn = repn
        self._node = node

    @property
    def name(self):
        return 'nl(' + self._node.name + ')'


class AMPLRepn:
    """The "compiled" representation of an expression in AMPL NL format.

    This stores a compiled form of an expression in the AMPL "NL"
    format.  The data structure contains 6 fields:

    Attributes
    ----------
    mult : float

        A constant multiplier applied to this expression.  The
        :py:class`AMPLRepn` returned by the :py:class`AMPLRepnVisitor`
        should always have `mult` == 1.

    const : float

        The constant portion of this expression

    linear : Dict[int, float] or None

        Mapping of `id(VarData)` to linear coefficient

    nonlinear : Tuple[str, List[int]] or List[Tuple[str, List[int]]] or None

        The general nonlinear portion of the compiled expression as a
        tuple of two parts:

           - the nl template string: this is the NL string with
             placeholders (``%s``) for all the variables that appear in
             the expression.

           - an iterable if the :class:`VarData` IDs that correspond to the
             placeholders in the nl template string

        This is `None` if there is no general nonlinear part of the
        expression.  Note that this can be a list of tuple fragments
        within AMPLRepnVisitor, but that list is concatenated to a
        single tuple when exiting the `AMPLRepnVisitor`.

    named_exprs : Set[int]

        A set of IDs point to named expressions (:py:class:`Expression`)
        objects appearing in this expression.

    nl : Tuple[str, Iterable[int]]

        This holds the complete compiled representation of this
        expression (including multiplier, constant, linear terms, and
        nonlinear fragment) using the same format as the `nonlinear`
        attribute.  This field (if not None) should be considered
        authoritative, as there are NL fragments that are not
        representable by {mult, const, linear, nonlinear} (e.g., string
        arguments).

    """

    __slots__ = ('nl', 'mult', 'const', 'linear', 'nonlinear', 'named_exprs')

    template = TextNLTemplate

    def __init__(self, const, linear, nonlinear):
        self.nl = None
        self.mult = 1
        self.const = const
        self.linear = linear
        if nonlinear is None:
            self.nonlinear = self.named_exprs = None
        else:
            nl, nl_args, self.named_exprs = nonlinear
            self.nonlinear = nl, nl_args

    def __str__(self):
        return (
            f'AMPLRepn(mult={self.mult}, const={self.const}, '
            f'linear={self.linear}, nonlinear={self.nonlinear}, '
            f'nl={self.nl}, named_exprs={self.named_exprs})'
        )

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return isinstance(other.__class__, AMPLRepn) and (
            self.nl == other.nl
            and self.mult == other.mult
            and self.const == other.const
            and self.linear == other.linear
            and self.nonlinear == other.nonlinear
            and self.named_exprs == other.named_exprs
        )

    def __hash__(self):
        # Approximation of the Python default object hash
        # (4 LSB are rolled to the MSB to reduce hash collisions)
        return id(self) // 16 + (
            (id(self) & 15) << 8 * ctypes.sizeof(ctypes.c_void_p) - 4
        )

    def duplicate(self):
        ans = self.__class__.__new__(self.__class__)
        ans.nl = self.nl
        ans.mult = self.mult
        ans.const = self.const
        ans.linear = None if self.linear is None else dict(self.linear)
        ans.nonlinear = self.nonlinear
        ans.named_exprs = None if self.named_exprs is None else set(self.named_exprs)
        return ans

    def compile_repn(self, prefix='', args=None, named_exprs=None):
        template = self.template
        if self.mult != 1:
            if self.mult == -1:
                prefix += template.negation
            else:
                prefix += template.multiplier % self.mult
            self.mult = 1
        if self.named_exprs is not None:
            if named_exprs is None:
                named_exprs = set(self.named_exprs)
            else:
                named_exprs.update(self.named_exprs)
        if self.nl is not None:
            # This handles both named subexpressions and embedded
            # non-numeric (e.g., string) arguments.
            nl, nl_args = self.nl
            if prefix:
                nl = prefix + nl
            if args is not None:
                assert args is not nl_args
                args.extend(nl_args)
            else:
                args = list(nl_args)
            if nl_args:
                # For string arguments, nl_args is an empty tuple and
                # self.named_exprs is None.  For named subexpressions,
                # we are guaranteed that named_exprs is NOT None.  We
                # need to ensure that the named subexpression that we
                # are returning is added to the named_exprs set.
                named_exprs.update(nl_args)
            return nl, args, named_exprs

        if args is None:
            args = []
        if self.linear:
            nterms = -len(args)
            _v_template = template.var
            _m_template = template.monomial
            # Because we are compiling this expression (into a NL
            # expression), we will go ahead and filter the 0*x terms
            # from the expression.  Note that the args are accumulated
            # by side-effect, which prevents iterating over the linear
            # terms twice.
            nl_sum = ''.join(
                args.append(v) or (_v_template if c == 1 else _m_template % c)
                for v, c in self.linear.items()
                if c
            )
            nterms += len(args)
        else:
            nterms = 0
            nl_sum = ''
        if self.nonlinear:
            if self.nonlinear.__class__ is list:
                nterms += len(self.nonlinear)
                nl_sum += ''.join(map(itemgetter(0), self.nonlinear))
                deque(map(args.extend, map(itemgetter(1), self.nonlinear)), maxlen=0)
            else:
                nterms += 1
                nl_sum += self.nonlinear[0]
                args.extend(self.nonlinear[1])
        if self.const:
            nterms += 1
            nl_sum += template.const % self.const

        if nterms > 2:
            return (prefix + (template.nary_sum % nterms) + nl_sum, args, named_exprs)
        elif nterms == 2:
            return prefix + template.binary_sum + nl_sum, args, named_exprs
        elif nterms == 1:
            return prefix + nl_sum, args, named_exprs
        else:  # nterms == 0
            return prefix + (template.const % 0), args, named_exprs

    def compile_nonlinear_fragment(self):
        if not self.nonlinear:
            self.nonlinear = None
            return
        args = []
        nterms = len(self.nonlinear)
        nl_sum = ''.join(map(itemgetter(0), self.nonlinear))
        deque(map(args.extend, map(itemgetter(1), self.nonlinear)), maxlen=0)

        if nterms > 2:
            self.nonlinear = (self.template.nary_sum % nterms) + nl_sum, args
        elif nterms == 2:
            self.nonlinear = self.template.binary_sum + nl_sum, args
        else:  # nterms == 1:
            self.nonlinear = nl_sum, args

    def append(self, other):
        """Append a child result from acceptChildResult

        Notes
        -----
        This method assumes that the operator was "+". It is implemented
        so that we can directly use an AMPLRepn() as a data object in
        the expression walker (thereby avoiding the function call for a
        custom callback)

        """
        # Note that self.mult will always be 1 (we only call append()
        # within a sum, so there is no opportunity for self.mult to
        # change). Omitting the assertion for efficiency.
        # assert self.mult == 1
        _type = other[0]
        if _type is _MONOMIAL:
            _, v, c = other
            if v in self.linear:
                self.linear[v] += c
            else:
                self.linear[v] = c
        elif _type is _GENERAL:
            _, other = other
            if other.nl is not None and other.nl[1]:
                if other.linear:
                    # This is a named expression with both a linear and
                    # nonlinear component.  We want to merge it with
                    # this AMPLRepn, preserving the named expression for
                    # only the nonlinear component (merging the linear
                    # component with this AMPLRepn).
                    pass
                else:
                    # This is a nonlinear-only named expression,
                    # possibly with a multiplier that is not 1.  Compile
                    # it and append it (this both resolves the
                    # multiplier, and marks the named expression as
                    # having been used)
                    other = other.compile_repn('', None, self.named_exprs)
                    nl, nl_args, self.named_exprs = other
                    self.nonlinear.append((nl, nl_args))
                    return
            if other.named_exprs is not None:
                if self.named_exprs is None:
                    self.named_exprs = set(other.named_exprs)
                else:
                    self.named_exprs.update(other.named_exprs)
            if other.mult != 1:
                mult = other.mult
                self.const += mult * other.const
                if other.linear:
                    linear = self.linear
                    for v, c in other.linear.items():
                        if v in linear:
                            linear[v] += c * mult
                        else:
                            linear[v] = c * mult
                if other.nonlinear:
                    if other.nonlinear.__class__ is list:
                        other.compile_nonlinear_fragment()
                    if mult == -1:
                        prefix = self.template.negation
                    else:
                        prefix = self.template.multiplier % mult
                    self.nonlinear.append(
                        (prefix + other.nonlinear[0], other.nonlinear[1])
                    )
            else:
                self.const += other.const
                if other.linear:
                    linear = self.linear
                    for v, c in other.linear.items():
                        if v in linear:
                            linear[v] += c
                        else:
                            linear[v] = c
                if other.nonlinear:
                    if other.nonlinear.__class__ is list:
                        self.nonlinear.extend(other.nonlinear)
                    else:
                        self.nonlinear.append(other.nonlinear)
        elif _type is _CONSTANT:
            self.const += other[1]

    def to_expr(self, var_map):
        if self.nl is not None or self.nonlinear is not None:
            # TODO: support converting general nonlinear expressions
            # back to Pyomo expressions.  This will require an AMPL
            # parser.
            raise MouseTrap("Cannot convert nonlinear AMPLRepn to Pyomo Expression")
        if self.linear:
            # Explicitly generate the LinearExpression.  At time of
            # writing, this is about 40% faster than standard operator
            # overloading for O(1000) element sums
            ans = LinearExpression(
                [coef * var_map[vid] for vid, coef in self.linear.items()]
            )
            ans += self.const
        else:
            ans = self.const
        return ans * self.mult


class DebugAMPLRepn(AMPLRepn):
    """An `AMPLRepn` that uses the "debug" (annotated) NL format

    This is identical to the :py:class:`AMPLRepn` class, except it is
    built using the `TextNLDebugTemplate` formatting template.  This
    format includes descriptions of the operators and variable /
    expression names in the NL text.

    """

    __slots__ = ()
    template = TextNLDebugTemplate


def handle_negation_node(visitor, node, arg1):
    if arg1[0] is _MONOMIAL:
        return (_MONOMIAL, arg1[1], -1 * arg1[2])
    elif arg1[0] is _GENERAL:
        arg1[1].mult *= -1
        return arg1
    elif arg1[0] is _CONSTANT:
        return (_CONSTANT, -1 * arg1[1])
    else:
        raise RuntimeError("%s: %s" % (type(arg1[0]), arg1))


def handle_product_node(visitor, node, arg1, arg2):
    if arg2[0] is _CONSTANT:
        arg2, arg1 = arg1, arg2
    if arg1[0] is _CONSTANT:
        mult = arg1[1]
        if not mult:
            # simplify multiplication by 0 (if arg2 is zero, the
            # simplification happens when we evaluate the constant
            # below).  Note that this is not IEEE-754 compliant, and
            # will map 0*inf and 0*nan to 0 (and not to nan).  We are
            # including this for backwards compatibility with the NLv1
            # writer, but arguably we should deprecate/remove this
            # "feature" in the future.
            if arg2[0] is _CONSTANT:
                _prod = mult * arg2[1]
                if _prod:
                    deprecation_warning(
                        f"Encountered {mult}*{_inv2str(arg2[1])} in expression tree.  "
                        "Mapping the NaN result to 0 for compatibility "
                        "with the nl_v1 writer.  In the future, this NaN "
                        "will be preserved/emitted to comply with IEEE-754.",
                        version='6.4.3',
                    )
                    _prod = 0
                return (_CONSTANT, _prod)
            return arg1
        if mult == 1:
            return arg2
        elif arg2[0] is _MONOMIAL:
            if mult != mult:
                # This catches mult (i.e., arg1) == nan
                return arg1
            return (_MONOMIAL, arg2[1], mult * arg2[2])
        elif arg2[0] is _GENERAL:
            if mult != mult:
                # This catches mult (i.e., arg1) == nan
                return arg1
            arg2[1].mult *= mult
            return arg2
        elif arg2[0] is _CONSTANT:
            if not arg2[1]:
                # Simplify multiplication by 0; see note above about
                # IEEE-754 incompatibility.
                _prod = mult * arg2[1]
                if _prod:
                    deprecation_warning(
                        f"Encountered {_inv2str(mult)}*{arg2[1]} in expression tree.  "
                        "Mapping the NaN result to 0 for compatibility "
                        "with the nl_v1 writer.  In the future, this NaN "
                        "will be preserved/emitted to comply with IEEE-754.",
                        version='6.4.3',
                    )
                    _prod = 0
                return (_CONSTANT, _prod)
            return (_CONSTANT, mult * arg2[1])
    nonlin = visitor.node_result_to_amplrepn(arg1).compile_repn(
        visitor.template.product
    )
    nonlin = visitor.node_result_to_amplrepn(arg2).compile_repn(*nonlin)
    return (_GENERAL, visitor.Result(0, None, nonlin))


def handle_division_node(visitor, node, arg1, arg2):
    if arg2[0] is _CONSTANT:
        div = arg2[1]
        if div == 1:
            return arg1
        if arg1[0] is _MONOMIAL:
            tmp = apply_node_operation(node, (arg1[2], div))
            if tmp != tmp:
                # This catches if the coefficient division results in nan
                return _CONSTANT, tmp
            return (_MONOMIAL, arg1[1], tmp)
        elif arg1[0] is _GENERAL:
            tmp = apply_node_operation(node, (arg1[1].mult, div))
            if tmp != tmp:
                # This catches if the multiplier division results in nan
                return _CONSTANT, tmp
            arg1[1].mult = tmp
            return arg1
        elif arg1[0] is _CONSTANT:
            return _CONSTANT, apply_node_operation(node, (arg1[1], div))
    elif arg1[0] is _CONSTANT and not arg1[1]:
        return _CONSTANT, 0
    nonlin = visitor.node_result_to_amplrepn(arg1).compile_repn(
        visitor.template.division
    )
    nonlin = visitor.node_result_to_amplrepn(arg2).compile_repn(*nonlin)
    return (_GENERAL, visitor.Result(0, None, nonlin))


def handle_pow_node(visitor, node, arg1, arg2):
    if arg2[0] is _CONSTANT:
        if arg1[0] is _CONSTANT:
            ans = apply_node_operation(node, (arg1[1], arg2[1]))
            if ans.__class__ in native_complex_types:
                ans = complex_number_error(ans, visitor, node)
            return _CONSTANT, ans
        elif not arg2[1]:
            return _CONSTANT, 1
        elif arg2[1] == 1:
            return arg1
    nonlin = visitor.node_result_to_amplrepn(arg1).compile_repn(visitor.template.pow)
    nonlin = visitor.node_result_to_amplrepn(arg2).compile_repn(*nonlin)
    return (_GENERAL, visitor.Result(0, None, nonlin))


def handle_abs_node(visitor, node, arg1):
    if arg1[0] is _CONSTANT:
        return (_CONSTANT, abs(arg1[1]))
    nonlin = visitor.node_result_to_amplrepn(arg1).compile_repn(visitor.template.abs)
    return (_GENERAL, visitor.Result(0, None, nonlin))


def handle_unary_node(visitor, node, arg1):
    if arg1[0] is _CONSTANT:
        return _CONSTANT, apply_node_operation(node, (arg1[1],))
    nonlin = visitor.node_result_to_amplrepn(arg1).compile_repn(
        visitor.template.unary[node.name]
    )
    return (_GENERAL, visitor.Result(0, None, nonlin))


def handle_exprif_node(visitor, node, arg1, arg2, arg3):
    if arg1[0] is _CONSTANT:
        if arg1[1]:
            return arg2
        else:
            return arg3
    nonlin = visitor.node_result_to_amplrepn(arg1).compile_repn(visitor.template.exprif)
    nonlin = visitor.node_result_to_amplrepn(arg2).compile_repn(*nonlin)
    nonlin = visitor.node_result_to_amplrepn(arg3).compile_repn(*nonlin)
    return (_GENERAL, visitor.Result(0, None, nonlin))


def handle_equality_node(visitor, node, arg1, arg2):
    if arg1[0] is _CONSTANT and arg2[0] is _CONSTANT:
        return (_CONSTANT, arg1[1] == arg2[1])
    nonlin = visitor.node_result_to_amplrepn(arg1).compile_repn(
        visitor.template.equality
    )
    nonlin = visitor.node_result_to_amplrepn(arg2).compile_repn(*nonlin)
    return (_GENERAL, visitor.Result(0, None, nonlin))


def handle_inequality_node(visitor, node, arg1, arg2):
    if arg1[0] is _CONSTANT and arg2[0] is _CONSTANT:
        return (_CONSTANT, node._apply_operation((arg1[1], arg2[1])))
    nonlin = visitor.node_result_to_amplrepn(arg1).compile_repn(
        visitor.template.strict_inequality_map[node.strict]
    )
    nonlin = visitor.node_result_to_amplrepn(arg2).compile_repn(*nonlin)
    return (_GENERAL, visitor.Result(0, None, nonlin))


def handle_ranged_inequality_node(visitor, node, arg1, arg2, arg3):
    if arg1[0] is _CONSTANT and arg2[0] is _CONSTANT and arg3[0] is _CONSTANT:
        return (_CONSTANT, node._apply_operation((arg1[1], arg2[1], arg3[1])))
    op = visitor.template.strict_inequality_map[node.strict]
    nl, args, named = visitor.node_result_to_amplrepn(arg1).compile_repn(
        visitor.template.and_expr + op[0]
    )
    nl2, args2, named = visitor.node_result_to_amplrepn(arg2).compile_repn(
        '', None, named
    )
    nl += nl2 + op[1] + nl2
    args.extend(args2)
    args.extend(args2)
    nonlin = visitor.node_result_to_amplrepn(arg3).compile_repn(nl, args, named)
    return (_GENERAL, visitor.Result(0, None, nonlin))


def handle_named_expression_node(visitor, node, arg1):
    _id = id(node)
    # Note that while named subexpressions ('defined variables' in the
    # ASL NL file vernacular) look like variables, they are not allowed
    # to appear in the 'linear' portion of a constraint / objective
    # definition.  We will return this as a "var" template, but
    # wrapped in the nonlinear portion of the expression tree.
    repn = visitor.node_result_to_amplrepn(arg1)

    # A local copy of the expression source list.  This will be updated
    # later if the same Expression node is encountered in another
    # expression tree.
    #
    # This is a 3-tuple [con_id, obj_id, substitute_expression].  If the
    # expression is used by more than 1 constraint / objective, then the
    # id is set to 0.  If it is not used by any, then it is None.
    # substitute_expression is a bool indicating if this named
    # subexpression tree should be directly substituted into any
    # expression tree that references this node (i.e., do NOT emit the V
    # line).
    expression_source = [None, None, False]
    # Record this common expression
    visitor.subexpression_cache[_id] = (
        # 0: the "component" that generated this expression ID
        node,
        # 1: the common subexpression (to be written out)
        repn,
        # 2: the source usage information for this subexpression:
        #    [(con_id, obj_id, substitute); see above]
        expression_source,
    )

    # As we will eventually need the compiled form of any nonlinear
    # expression, we will go ahead and compile it here.  We do not
    # do the same for the linear component as we will only need the
    # linear component compiled to a dict if we are emitting the
    # original (linear + nonlinear) V line (which will not happen if
    # the V line is part of a larger linear operator).
    if repn.nonlinear.__class__ is list:
        repn.compile_nonlinear_fragment()

    if not visitor.use_named_exprs:
        return _GENERAL, repn.duplicate()

    mult, repn.mult = repn.mult, 1
    if repn.named_exprs is None:
        repn.named_exprs = set()

    # When converting this shared subexpression to a (nonlinear)
    # node, we want to just reference this subexpression:
    repn.nl = (visitor.template.var, (_id,))

    if repn.nonlinear:
        if repn.linear:
            # If this expression has both linear and nonlinear
            # components, we will follow the ASL convention and break
            # the named subexpression into two named subexpressions: one
            # that is only the nonlinear component and one that has the
            # const/linear component (and references the first).  This
            # will allow us to propagate linear coefficients up from
            # named subexpressions when appropriate.
            sub_node = NLFragment(repn, node)
            sub_id = id(sub_node)
            sub_repn = visitor.Result(0, None, None)
            sub_repn.nonlinear = repn.nonlinear
            sub_repn.nl = (visitor.template.var, (sub_id,))
            sub_repn.named_exprs = set(repn.named_exprs)

            repn.named_exprs.add(sub_id)
            repn.nonlinear = sub_repn.nl

            # See above for the meaning of this source information
            nl_info = list(expression_source)
            visitor.subexpression_cache[sub_id] = (sub_node, sub_repn, nl_info)
            # It is important that the NL subexpression comes before the
            # main named expression: re-insert the original named
            # expression (so that the nonlinear sub_node comes first
            # when iterating over subexpression_cache)
            visitor.subexpression_cache[_id] = visitor.subexpression_cache.pop(_id)
        else:
            nl_info = expression_source
    else:
        repn.nonlinear = None
        if repn.linear:
            if (
                not repn.const
                and len(repn.linear) == 1
                and next(iter(repn.linear.values())) == 1
            ):
                # This Expression holds only a variable (multiplied by
                # 1).  Do not emit this as a named variable and instead
                # just inject the variable where this expression is
                # used.
                repn.nl = None
                expression_source[2] = True
        else:
            # This Expression holds only a constant.  Do not emit this
            # as a named variable and instead just inject the constant
            # where this expression is used.
            repn.nl = None
            expression_source[2] = True

    if mult != 1:
        repn.const *= mult
        if repn.linear:
            _lin = repn.linear
            for v in repn.linear:
                _lin[v] *= mult
        if repn.nonlinear:
            if mult == -1:
                prefix = visitor.template.negation
            else:
                prefix = visitor.template.multiplier % mult
            repn.nonlinear = prefix + repn.nonlinear[0], repn.nonlinear[1]

    if expression_source[2]:
        if repn.linear:
            assert len(repn.linear) == 1 and not repn.const
            return (_MONOMIAL,) + next(iter(repn.linear.items()))
        else:
            return (_CONSTANT, repn.const)

    return (_GENERAL, repn.duplicate())


def handle_external_function_node(visitor, node, *args):
    func = node._fcn._function
    # There is a special case for external functions: these are the only
    # expressions that can accept string arguments. As we currently pass
    # these as 'precompiled' GENERAL AMPLRepns, the normal trap for
    # constant subexpressions will miss string arguments.  We will catch
    # that case here by looking for NL fragments with no variable
    # references.  Note that the NL fragment is NOT the raw string
    # argument that we want to evaluate: the raw string is in the
    # `const` field.
    if all(
        arg[0] is _CONSTANT or (arg[0] is _GENERAL and arg[1].nl and not arg[1].nl[1])
        for arg in args
    ):
        arg_list = [arg[1] if arg[0] is _CONSTANT else arg[1].const for arg in args]
        return _CONSTANT, apply_node_operation(node, arg_list)
    if func in visitor.external_functions:
        if node._fcn._library != visitor.external_functions[func][1]._library:
            raise RuntimeError(
                "The same external function name (%s) is associated "
                "with two different libraries (%s through %s, and %s "
                "through %s).  The ASL solver will fail to link "
                "correctly."
                % (
                    func,
                    visitor.external_functions[func]._library,
                    visitor.external_functions[func]._library.name,
                    node._fcn._library,
                    node._fcn.name,
                )
            )
    else:
        visitor.external_functions[func] = (len(visitor.external_functions), node._fcn)
    comment = f'\t#{node.local_name}' if visitor.symbolic_solver_labels else ''
    nl = visitor.template.external_fcn % (
        visitor.external_functions[func][0],
        len(args),
        comment,
    )
    arg_ids = []
    named_exprs = set()
    for arg in args:
        _id = id(arg)
        arg_ids.append(_id)
        visitor.subexpression_cache[_id] = (
            arg,
            visitor.Result(
                0,
                None,
                visitor.node_result_to_amplrepn(arg).compile_repn(
                    named_exprs=named_exprs
                ),
            ),
            (None, None, True),
        )
    if not named_exprs:
        named_exprs = None
    return (
        _GENERAL,
        visitor.Result(0, None, (nl + '%s' * len(arg_ids), arg_ids, named_exprs)),
    )


_operator_handles = ExitNodeDispatcher(
    {
        NegationExpression: handle_negation_node,
        ProductExpression: handle_product_node,
        DivisionExpression: handle_division_node,
        PowExpression: handle_pow_node,
        AbsExpression: handle_abs_node,
        UnaryFunctionExpression: handle_unary_node,
        Expr_ifExpression: handle_exprif_node,
        EqualityExpression: handle_equality_node,
        InequalityExpression: handle_inequality_node,
        RangedExpression: handle_ranged_inequality_node,
        Expression: handle_named_expression_node,
        ExternalFunctionExpression: handle_external_function_node,
        # These are handled explicitly in beforeChild():
        # LinearExpression: handle_linear_expression,
        # SumExpression: handle_sum_expression,
        #
        # Note: MonomialTermExpression is only hit when processing NPV
        # subexpressions that raise errors (e.g., log(0) * m.x), so no
        # special processing is needed [it is just a product expression]
        MonomialTermExpression: handle_product_node,
    }
)


class AMPLBeforeChildDispatcher(BeforeChildDispatcher):
    __slots__ = ()

    def __init__(self):
        # Special linear / summation expressions
        self[MonomialTermExpression] = self._before_monomial
        self[LinearExpression] = self._before_linear
        self[SumExpression] = self._before_general_expression

    @staticmethod
    def _record_var(visitor, var):
        # We always add all indices to the var_map at once so that
        # we can honor deterministic ordering of unordered sets
        # (because the user could have iterated over an unordered
        # set when constructing an expression, thereby altering the
        # order in which we would see the variables)
        vm = visitor.var_map
        try:
            _iter = var.parent_component().values(visitor.sorter)
        except AttributeError:
            # Note that this only works for the AML, as kernel does not
            # provide a parent_component()
            _iter = (var,)
        for v in _iter:
            if v.fixed:
                continue
            vm[id(v)] = v

    @staticmethod
    def _before_string(visitor, child):
        visitor.encountered_string_arguments = True
        ans = visitor.Result(child, None, None)
        ans.nl = (visitor.template.string % (len(child), child), ())
        return False, (_GENERAL, ans)

    @staticmethod
    def _before_var(visitor, child):
        _id = id(child)
        if _id not in visitor.var_map:
            if child.fixed:
                if _id not in visitor.fixed_vars:
                    visitor.cache_fixed_var(_id, child)
                return False, (_CONSTANT, visitor.fixed_vars[_id])
            _before_child_handlers._record_var(visitor, child)
        return False, (_MONOMIAL, _id, 1)

    @staticmethod
    def _before_monomial(visitor, child):
        #
        # The following are performance optimizations for common
        # situations (Monomial terms and Linear expressions)
        #
        arg1, arg2 = child._args_
        if arg1.__class__ not in native_types:
            try:
                arg1 = check_constant(visitor.evaluate(arg1), arg1, visitor)
            except (ValueError, ArithmeticError):
                return True, None

        # Trap multiplication by 0 and nan.
        if not arg1:
            if arg2.fixed:
                _id = id(arg2)
                if _id not in visitor.fixed_vars:
                    visitor.cache_fixed_var(id(arg2), arg2)
                arg2 = visitor.fixed_vars[_id]
                if arg2 != arg2:
                    deprecation_warning(
                        f"Encountered {arg1}*{_inv2str(arg2)} in expression tree.  "
                        "Mapping the NaN result to 0 for compatibility "
                        "with the nl_v1 writer.  In the future, this NaN "
                        "will be preserved/emitted to comply with IEEE-754.",
                        version='6.4.3',
                    )
            return False, (_CONSTANT, arg1)

        _id = id(arg2)
        if _id not in visitor.var_map:
            if arg2.fixed:
                if _id not in visitor.fixed_vars:
                    visitor.cache_fixed_var(_id, arg2)
                return False, (_CONSTANT, arg1 * visitor.fixed_vars[_id])
            _before_child_handlers._record_var(visitor, arg2)
        return False, (_MONOMIAL, _id, arg1)

    @staticmethod
    def _before_linear(visitor, child):
        # Because we are going to modify the LinearExpression in this
        # walker, we need to make a copy of the arg list from the original
        # expression tree.
        var_map = visitor.var_map
        const = 0
        linear = {}
        for arg in child.args:
            if arg.__class__ is MonomialTermExpression:
                arg1, arg2 = arg._args_
                if arg1.__class__ not in native_types:
                    try:
                        arg1 = check_constant(visitor.evaluate(arg1), arg1, visitor)
                    except (ValueError, ArithmeticError):
                        return True, None

                # Trap multiplication by 0 and nan.
                if not arg1:
                    if arg2.fixed:
                        arg2 = check_constant(arg2.value, arg2, visitor)
                        if arg2 != arg2:
                            deprecation_warning(
                                f"Encountered {arg1}*{_inv2str(arg2)} in expression "
                                "tree.  Mapping the NaN result to 0 for compatibility "
                                "with the nl_v1 writer.  In the future, this NaN "
                                "will be preserved/emitted to comply with IEEE-754.",
                                version='6.4.3',
                            )
                    continue

                _id = id(arg2)
                if _id not in var_map:
                    if arg2.fixed:
                        if _id not in visitor.fixed_vars:
                            visitor.cache_fixed_var(_id, arg2)
                        const += arg1 * visitor.fixed_vars[_id]
                        continue
                    _before_child_handlers._record_var(visitor, arg2)
                    linear[_id] = arg1
                elif _id in linear:
                    linear[_id] += arg1
                else:
                    linear[_id] = arg1
            elif arg.__class__ in native_types:
                const += arg
            elif arg.is_variable_type():
                _id = id(arg)
                if _id not in var_map:
                    if arg.fixed:
                        if _id not in visitor.fixed_vars:
                            visitor.cache_fixed_var(_id, arg)
                        const += visitor.fixed_vars[_id]
                        continue
                    _before_child_handlers._record_var(visitor, arg)
                    linear[_id] = 1
                elif _id in linear:
                    linear[_id] += 1
                else:
                    linear[_id] = 1
            else:
                try:
                    const += check_constant(visitor.evaluate(arg), arg, visitor)
                except (ValueError, ArithmeticError):
                    return True, None

        if linear:
            return False, (_GENERAL, visitor.Result(const, linear, None))
        else:
            return False, (_CONSTANT, const)

    @staticmethod
    def _before_named_expression(visitor, child):
        _id = id(child)
        if _id in visitor.subexpression_cache:
            obj, repn, info = visitor.subexpression_cache[_id]
            if info[2]:
                if repn.linear:
                    return False, (_MONOMIAL, next(iter(repn.linear)), 1)
                else:
                    return False, (_CONSTANT, repn.const)
            return False, (_GENERAL, repn.duplicate())
        else:
            return True, None


_before_child_handlers = AMPLBeforeChildDispatcher()


class AMPLRepnVisitor(StreamBasedExpressionVisitor):
    def __init__(
        self,
        subexpression_cache,
        external_functions,
        var_map,
        used_named_expressions,
        symbolic_solver_labels,
        use_named_exprs,
        sorter,
    ):
        super().__init__()
        self.subexpression_cache = subexpression_cache
        self.external_functions = external_functions
        self.active_expression_source = None
        self.var_map = var_map
        self.used_named_expressions = used_named_expressions
        self.symbolic_solver_labels = symbolic_solver_labels
        self.use_named_exprs = use_named_exprs
        self.encountered_string_arguments = False
        self.fixed_vars = {}
        self._eval_expr_visitor = _EvaluationVisitor(True)
        self.evaluate = self._eval_expr_visitor.dfs_postorder_stack
        self.sorter = sorter

        if symbolic_solver_labels:
            self.Result = DebugAMPLRepn
        else:
            self.Result = AMPLRepn
        self.template = self.Result.template

    def cache_fixed_var(self, _id, child):
        val = check_constant(child.value, child, self)
        lb, ub = child.bounds
        if (lb is not None and lb - val > TOL) or (ub is not None and ub - val < -TOL):
            raise InfeasibleConstraintException(
                "model contains a trivially infeasible "
                f"variable '{child.name}' (fixed value "
                f"{val} outside bounds [{lb}, {ub}])."
            )
        self.fixed_vars[_id] = check_constant(child.value, child, self)

    def node_result_to_amplrepn(self, data):
        if data[0] is _GENERAL:
            return data[1]
        elif data[0] is _MONOMIAL:
            _, v, c = data
            if c:
                return self.Result(0, {v: c}, None)
            else:
                return self.Result(0, None, None)
        elif data[0] is _CONSTANT:
            return self.Result(data[1], None, None)
        else:
            raise DeveloperError("unknown result type")

    def initializeWalker(self, expr):
        expr, src, src_idx, self.expression_scaling_factor = expr
        self.active_expression_source = (src_idx, id(src))
        walk, result = self.beforeChild(None, expr, 0)
        if not walk:
            return False, self.finalizeResult(result)
        return True, expr

    def beforeChild(self, node, child, child_idx):
        return _before_child_handlers[child.__class__](self, child)

    def enterNode(self, node):
        # SumExpression are potentially large nary operators.  Directly
        # populate the result
        if node.__class__ in sum_like_expression_types:
            data = self.Result(0, {}, None)
            data.nonlinear = []
            return node.args, data
        else:
            return node.args, []

    def exitNode(self, node, data):
        if data.__class__ is self.Result:
            # If the summation resulted in a constant, return the constant
            if data.linear or data.nonlinear or data.nl:
                return (_GENERAL, data)
            else:
                return (_CONSTANT, data.const)
        #
        # General expressions...
        #
        return _operator_handles[node.__class__](self, node, *data)

    def finalizeResult(self, result):
        ans = self.node_result_to_amplrepn(result)

        # Multiply the expression by the scaling factor provided by the caller
        ans.mult *= self.expression_scaling_factor

        # If this was a nonlinear named expression, and that expression
        # has no linear portion, then we will directly use this as a
        # named expression.  We need to mark that the expression was
        # used and return it as a simple nonlinear expression pointing
        # to this named expression.  In all other cases, we will return
        # the processed representation (which will reference the
        # nonlinear-only named subexpression - if it exists - but not
        # this outer named expression).  This prevents accidentally
        # recharacterizing variables that only appear linearly as
        # nonlinear variables.
        if ans.nl is not None:
            if not ans.nl[1]:
                raise ValueError("Numeric expression resolved to a string constant")
            # This *is* a named subexpression.  If there is no linear
            # component, then replace this expression with the named
            # expression.  The mult will be handled later.  We know that
            # the const is built into the nonlinear expression, because
            # it cannot be changed "in place" (only through addition,
            # which would have "cleared" the nl attribute)
            if not ans.linear:
                ans.named_exprs.update(ans.nl[1])
                ans.nonlinear = ans.nl
                ans.const = 0
            else:
                # This named expression has both a linear and a
                # nonlinear component, and possibly a multiplier and
                # constant.  We will not include this named expression
                # and instead will expose the components so that linear
                # variables are not accidentally re-characterized as
                # nonlinear.
                pass
            ans.nl = None

        if ans.nonlinear.__class__ is list:
            ans.compile_nonlinear_fragment()

        if not ans.linear:
            ans.linear = {}
        if ans.mult != 1:
            linear = ans.linear
            mult, ans.mult = ans.mult, 1
            ans.const *= mult
            if linear:
                for k in linear:
                    linear[k] *= mult
            if ans.nonlinear:
                if mult == -1:
                    prefix = self.template.negation
                else:
                    prefix = self.template.multiplier % mult
                ans.nonlinear = prefix + ans.nonlinear[0], ans.nonlinear[1]
        #
        self.active_expression_source = None
        return ans


def evaluate_ampl_nl_expression(nl, external_functions):
    expr = nl.splitlines()
    stack = []
    while expr:
        line = expr.pop()
        tokens = line.split()
        # remove tokens after the first comment
        for i, t in enumerate(tokens):
            if t.startswith('#'):
                tokens = tokens[:i]
                break
        if len(tokens) != 1:
            # skip blank lines
            if not tokens:
                continue
            if tokens[0][0] == 'f':
                # external function
                fid, nargs = tokens
                fid = int(fid[1:])
                nargs = int(nargs)
                fcn_id, ef = external_functions[fid]
                assert fid == fcn_id
                stack.append(ef.evaluate(tuple(stack.pop() for i in range(nargs))))
                continue
            raise DeveloperError(
                f"Unsupported line format _evaluate_constant_nl() "
                f"(we expect each line to contain a single token): '{line}'"
            )
        term = tokens[0]
        # the "command" can be determined by the first character on the line
        cmd = term[0]
        # Note that we will unpack the line into the expected number of
        # explicit arguments as a form of error checking
        if cmd == 'n':
            # numeric constant
            stack.append(float(term[1:]))
        elif cmd == 'o':
            # operator
            nargs, fcn = nl_operators[int(term[1:])]
            if nargs is None:
                nargs = int(stack.pop())
            stack.append(fcn(*(stack.pop() for i in range(nargs))))
        elif cmd in '1234567890':
            # this is either a single int (e.g., the nargs in a nary
            # sum) or a string argument.  Preserve it as-is until later
            # when we know which we are expecting.
            stack.append(term)
        elif cmd == 'h':
            stack.append(term.split(':', 1)[1])
        else:
            raise DeveloperError(
                f"Unsupported NL operator in _evaluate_constant_nl(): '{line}'"
            )
    assert len(stack) == 1
    return stack[0]



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#
# Test the path solver
#

import os

import pyomo.common.unittest as unittest

from pyomo.common.dependencies import yaml, yaml_available, yaml_load_args
from pyomo.common.fileutils import this_file_dir, PYOMO_ROOT_DIR
import pyomo.opt
import pyomo.scripting.pyomo_main as pyomo_main
from pyomo.scripting.util import cleanup

# Ensure plugins are registered
import pyomo.environ

currdir = this_file_dir()
exdir = os.path.join(PYOMO_ROOT_DIR, 'examples', 'mpec')

solvers = pyomo.opt.check_available_solvers('path')


class CommonTests:
    solve = True
    solver = 'path'

    def run_solver(self, *_args, **kwds):
        if self.solve:
            args = ['solve']
            args.append('--solver=' + self.solver)
            args.append('--save-results=result.yml')
            args.append('--results-format=yaml')
            args.append('--solver-options="lemke_start=automatic output_options=yes"')
        else:
            args = ['convert']
        args.append('-c')
        args.append('--symbolic-solver-labels')
        args.append('--file-determinism=2')

        if False:
            args.append('--stream-solver')
            args.append('--tempdir=' + currdir)
            args.append('--keepfiles')
            args.append('--logging=debug')

        args = args + list(_args)
        os.chdir(currdir)

        print('***')
        print(' '.join(args))
        try:
            output = pyomo_main.main(args)
        except SystemExit:
            output = None
        except:
            output = None
            raise
        cleanup()
        print('***')
        return output

    def referenceFile(self, problem, solver):
        return os.path.join(currdir, problem + '.txt')

    def getObjective(self, fname):
        FILE = open(fname, 'r')
        data = yaml.load(FILE, **yaml_load_args)
        FILE.close()
        solutions = data.get('Solution', [])
        ans = []
        for x in solutions:
            ans.append(x.get('Objective', {}))
        return ans

    def updateDocStrings(self):
        for key in dir(self):
            if key.startswith('test'):
                getattr(self, key).__doc__ = " (%s)" % getattr(self, key).__name__

    def test_munson1a(self):
        self.problem = 'test_munson1a'
        self.run_solver(os.path.join(exdir, 'munson1a.py'))
        self.check('munson1a', self.solver)

    def test_munson1b(self):
        self.problem = 'test_munson1b'
        self.run_solver(os.path.join(exdir, 'munson1b.py'))
        self.check('munson1b', self.solver)

    def test_munson1c(self):
        self.problem = 'test_munson1c'
        self.run_solver(os.path.join(exdir, 'munson1c.py'))
        self.check('munson1c', self.solver)

    def test_munson1d(self):
        self.problem = 'test_munson1d'
        self.run_solver(os.path.join(exdir, 'munson1d.py'))
        self.check('munson1d', self.solver)

    def check(self, problem, solver):
        refObj = self.getObjective(self.referenceFile(problem, solver))
        ansObj = self.getObjective(os.path.join(currdir, 'result.yml'))
        self.assertEqual(len(refObj), len(ansObj))
        for i in range(len(refObj)):
            self.assertEqual(len(refObj[i]), len(ansObj[i]))
            if isinstance(refObj[i], str):
                continue
            for key, val in refObj[i].items():
                self.assertAlmostEqual(
                    val['Value'], ansObj[i].get(key, None)['Value'], places=2
                )


@unittest.skipIf(not yaml_available, "YAML is not available")
@unittest.skipIf(not 'path' in solvers, "The 'path' executable is not available")
class Solve_PATH(unittest.TestCase, CommonTests):
    def tearDown(self):
        if os.path.exists(os.path.join(currdir, 'result.yml')):
            os.remove(os.path.join(currdir, 'result.yml'))


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# Routines to collect data in a structured format

from pyomo.common.collections import Bunch
from pyomo.core.base import Var, Constraint, Objective, maximize, minimize
from pyomo.repn.standard_repn import generate_standard_repn


def collect_linear_terms(block, unfixed):
    #
    # Variables are constraints of block
    # Constraints are unfixed variables of block and the parent model.
    #
    vnames = set()
    for obj in block.component_objects(Constraint, active=True):
        vnames.add(
            (obj.getname(fully_qualified=True, relative_to=block), obj.is_indexed())
        )
    cnames = set(unfixed)
    for obj in block.component_objects(Var, active=True):
        cnames.add(
            (obj.getname(fully_qualified=True, relative_to=block), obj.is_indexed())
        )
    #
    A = {}
    b_coef = {}
    c_rhs = {}
    c_sense = {}
    d_sense = None
    v_domain = {}
    #
    # Collect objective
    #
    for odata in block.component_objects(Objective, active=True):
        for ndx in odata:
            if odata[ndx].sense == maximize:
                o_terms = generate_standard_repn(
                    -1 * odata[ndx].expr, compute_values=False
                )
                d_sense = minimize
            else:
                o_terms = generate_standard_repn(odata[ndx].expr, compute_values=False)
                d_sense = maximize
            for var, coef in zip(o_terms.linear_vars, o_terms.linear_coefs):
                c_rhs[var.parent_component().local_name, var.index()] = coef
        # Stop after the first objective
        break
    #
    # Collect constraints
    #
    for data in block.component_objects(Constraint, active=True):
        name = data.getname(relative_to=block)
        for ndx in data:
            con = data[ndx]
            body_terms = generate_standard_repn(con.body, compute_values=False)
            if body_terms.is_fixed():
                #
                # If a constraint has a fixed body, then don't collect it.
                #
                continue
            lower_terms = (
                generate_standard_repn(con.lower, compute_values=False)
                if not con.lower is None
                else None
            )
            upper_terms = (
                generate_standard_repn(con.upper, compute_values=False)
                if not con.upper is None
                else None
            )
            #
            if not lower_terms is None and not lower_terms.is_constant():
                raise (
                    RuntimeError,
                    "Error during dualization:  Constraint '%s' has a lower bound that is non-constant",
                )
            if not upper_terms is None and not upper_terms.is_constant():
                raise (
                    RuntimeError,
                    "Error during dualization:  Constraint '%s' has an upper bound that is non-constant",
                )
            #
            for var, coef in zip(body_terms.linear_vars, body_terms.linear_coefs):
                try:
                    # The variable is in the subproblem
                    varname = var.parent_component().getname(
                        fully_qualified=True, relative_to=block
                    )
                except:
                    # The variable is somewhere else in the model
                    varname = var.parent_component().getname(
                        fully_qualified=True, relative_to=block.model()
                    )
                varndx = var.index()
                A.setdefault(varname, {}).setdefault(varndx, []).append(
                    Bunch(coef=coef, var=name, ndx=ndx)
                )
            #
            if not con.equality:
                #
                # Inequality constraint
                #
                if lower_terms is None:
                    #
                    # body <= upper
                    #
                    v_domain[name, ndx] = -1
                    b_coef[name, ndx] = upper_terms.constant - body_terms.constant
                elif upper_terms is None:
                    #
                    # lower <= body
                    #
                    v_domain[name, ndx] = 1
                    b_coef[name, ndx] = lower_terms.constant - body_terms.constant
                else:
                    #
                    # lower <= body <= upper
                    #
                    # Dual for lower bound
                    #
                    ndx_ = tuple(list(ndx).append('lb'))
                    v_domain[name, ndx_] = 1
                    b_coef[name, ndx] = lower_terms.constant - body_terms.constant
                    #
                    # Dual for upper bound
                    #
                    ndx_ = tuple(list(ndx).append('ub'))
                    v_domain[name, ndx_] = -1
                    b_coef[name, ndx] = upper_terms.constant - body_terms.constant
            else:
                #
                # Equality constraint
                #
                v_domain[name, ndx] = 0
                b_coef[name, ndx] = lower_terms.constant - body_terms.constant

    #
    # Collect bound constraints
    #
    def all_vars(b):
        """
        This conditionally chains together the active variables in the current block with
        the active variables in all of the parent blocks (if any exist).
        """
        for obj in b.component_objects(Var, active=True, descend_into=True):
            name = obj.parent_component().getname(fully_qualified=True, relative_to=b)
            yield (name, obj)
        #
        # Look through parent blocks
        #
        b = b.parent_block()
        while not b is None:
            for obj in b.component_objects(Var, active=True, descend_into=False):
                name = obj.parent_component().name
                yield (name, obj)
            b = b.parent_block()

    for name, data in all_vars(block):
        #
        # Skip fixed variables (in the parent)
        #
        if not (name, data.is_indexed()) in cnames:
            continue
        #
        # Iterate over all variable indices
        #
        for ndx in data:
            var = data[ndx]
            bounds = var.bounds
            if bounds[0] is None and bounds[1] is None:
                c_sense[name, ndx] = 'e'
            elif bounds[0] is None:
                if bounds[1] == 0.0:
                    c_sense[name, ndx] = 'g'
                else:
                    c_sense[name, ndx] = 'e'
                    #
                    # Add constraint that defines the upper bound
                    #
                    name_ = name + "_upper_"
                    varname = data.parent_component().getname(
                        fully_qualified=True, relative_to=block
                    )
                    varndx = data[ndx].index()
                    A.setdefault(varname, {}).setdefault(varndx, []).append(
                        Bunch(coef=1.0, var=name_, ndx=ndx)
                    )
                    #
                    v_domain[name_, ndx] = -1
                    b_coef[name_, ndx] = bounds[1]
            elif bounds[1] is None:
                if bounds[0] == 0.0:
                    c_sense[name, ndx] = 'l'
                else:
                    c_sense[name, ndx] = 'e'
                    #
                    # Add constraint that defines the lower bound
                    #
                    name_ = name + "_lower_"
                    varname = data.parent_component().getname(
                        fully_qualified=True, relative_to=block
                    )
                    varndx = data[ndx].index()
                    A.setdefault(varname, {}).setdefault(varndx, []).append(
                        Bunch(coef=1.0, var=name_, ndx=ndx)
                    )
                    #
                    v_domain[name_, ndx] = 1
                    b_coef[name_, ndx] = bounds[0]
            else:
                # Bounded above and below
                c_sense[name, ndx] = 'e'
                #
                # Add constraint that defines the upper bound
                #
                name_ = name + "_upper_"
                varname = data.parent_component().getname(
                    fully_qualified=True, relative_to=block
                )
                varndx = data[ndx].index()
                A.setdefault(varname, {}).setdefault(varndx, []).append(
                    Bunch(coef=1.0, var=name_, ndx=ndx)
                )
                #
                v_domain[name_, ndx] = -1
                b_coef[name_, ndx] = bounds[1]
                #
                # Add constraint that defines the lower bound
                #
                name_ = name + "_lower_"
                varname = data.parent_component().getname(
                    fully_qualified=True, relative_to=block
                )
                varndx = data[ndx].index()
                A.setdefault(varname, {}).setdefault(varndx, []).append(
                    Bunch(coef=1.0, var=name_, ndx=ndx)
                )
                #
                v_domain[name_, ndx] = 1
                b_coef[name_, ndx] = bounds[0]
    #
    return (A, b_coef, c_rhs, c_sense, d_sense, vnames, cnames, v_domain)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.common.dependencies import (
    matplotlib,
    matplotlib_available,
    scipy_available,
    networkx_available,
)
from pyomo.contrib.incidence_analysis.visualize import spy_dulmage_mendelsohn
from pyomo.contrib.incidence_analysis.tests.models_for_testing import (
    make_gas_expansion_model,
    make_dynamic_model,
    make_degenerate_solid_phase_model,
)


@unittest.skipUnless(matplotlib_available, "Matplotlib is not available")
@unittest.skipUnless(scipy_available, "SciPy is not available")
@unittest.skipUnless(networkx_available, "NetworkX is not available")
class TestSpy(unittest.TestCase):
    def test_spy_dulmage_mendelsohn(self):
        models = [
            make_gas_expansion_model(),
            make_dynamic_model(),
            make_degenerate_solid_phase_model(),
        ]
        for m in models:
            fig, ax = spy_dulmage_mendelsohn(m)
            # Note that this is a weak test. We just test that we can call the
            # plot method, it doesn't raise an error, and gives us back the
            # types we expect. We don't attempt to validate the resulting plot.
            self.assertTrue(isinstance(fig, matplotlib.pyplot.Figure))
            self.assertTrue(isinstance(ax, matplotlib.pyplot.Axes))


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.collections import ComponentSet
from pyomo.common.config import ConfigDict, ConfigValue
import pyomo.common.unittest as unittest
from pyomo.core import Block, ConcreteModel, Constraint, Objective, Var
from pyomo.util.config_domains import ComponentDataSet


def ComponentSetConfig():
    CONFIG = ConfigDict()
    CONFIG.declare(
        'var_set',
        ConfigValue(default=None, domain=ComponentDataSet(Var), doc="VarDataSet"),
    )
    CONFIG.declare(
        'var_and_constraint_set',
        ConfigValue(
            default=None,
            domain=ComponentDataSet(ctype=(Var, Constraint)),
            doc="VarAndConstraintSet",
        ),
    )
    return CONFIG


def a_model():
    m = ConcreteModel()
    m.x = Var()
    m.y = Var([1, 2])
    m.c = Constraint(expr=m.x + m.y[1] <= 3)
    m.c2 = Constraint(expr=m.y[2] >= 7)
    m.obj = Objective(expr=m.x + m.y[1] + m.y[2])
    m.b = Block()

    return m


class TestComponentDataSetDomain(unittest.TestCase):
    def test_var_set(self):
        m = a_model()
        config = ComponentSetConfig()
        self.assertIsNone(config.var_set)
        config.var_set = ComponentSet([m.x, m.y])
        self.assertIsInstance(config.var_set, ComponentSet)
        self.assertEqual(len(config.var_set), 3)
        for v in [m.x, m.y[1], m.y[2]]:
            self.assertIn(v, config.var_set)

        with self.assertRaisesRegex(
            ValueError,
            ".*Expected component or iterable of one "
            "of the following ctypes: Var.\n\t"
            "Received <class 'pyomo.core.base.constraint.ScalarConstraint'>",
        ):
            config.var_set = ComponentSet([m.y, m.c])

    def test_var_and_constraint_set(self):
        m = a_model()
        config = ComponentSetConfig()
        self.assertIsNone(config.var_and_constraint_set)
        config.var_and_constraint_set = ComponentSet([m.x, m.c])
        self.assertIsInstance(config.var_and_constraint_set, ComponentSet)
        self.assertEqual(len(config.var_and_constraint_set), 2)
        for v in [m.x, m.c]:
            self.assertIn(v, config.var_and_constraint_set)

        with self.assertRaisesRegex(
            ValueError,
            ".*Expected component or iterable of one "
            "of the following ctypes: Constraint, Var.\n\t"
            "Received <class 'pyomo.core.base.block.ScalarBlock'>",
        ):
            config.var_and_constraint_set = ComponentSet([m.y, m.c, m.b])

        with self.assertRaisesRegex(
            ValueError,
            ".*Expected component or iterable of one "
            "of the following ctypes: Constraint, Var.\n\t"
            "Received <class 'int'>",
        ):
            config.var_and_constraint_set = ComponentSet([3, m.y, m.c])

    def test_domain_name(self):
        config = ComponentSetConfig()
        self.assertEqual(config.get("var_set").domain_name(), "ComponentDataSet(Var)")
        self.assertEqual(
            config.get("var_and_constraint_set").domain_name(),
            "ComponentDataSet([Constraint, Var])",
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


from collections import OrderedDict, UserDict
from collections.abc import Mapping, MutableMapping, MutableSet, Sequence, Set

from .bunch import Bunch
from .component_map import ComponentMap, DefaultComponentMap
from .component_set import ComponentSet
from .orderedset import OrderedSet



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.expr.logical_expr import BooleanExpression


class PrecedenceExpression(BooleanExpression):
    PRECEDENCE = None

    def nargs(self):
        return 3

    @property
    def delay(self):
        return self._args_[2]

    def _to_string_impl(self, values, relation):
        delay = values[2]
        if delay == '0':
            first = values[0]
        elif delay[0] in '-+':
            first = "%s %s %s" % (values[0], delay[0], delay[1:])
        else:
            first = "%s + %s" % (values[0], delay)
        return "%s %s %s" % (first, relation, values[1])


class BeforeExpression(PrecedenceExpression):
    """
    Base class for all precedence expressions.

    args:
        args (tuple): child nodes of type IntervalVar. We expect them to be
                      (time_that_comes_before, time_that_comes_after, delay).
        delay: A (possibly negative) integer value representing the number of
               time periods delay in the precedence relationship
    """

    def _to_string(self, values, verbose, smap):
        return self._to_string_impl(values, "<=")


class AtExpression(PrecedenceExpression):
    """
    Base class for all precedence expressions.

    args:
        args (tuple): child nodes of type IntervalVar. We expect them to be
                      (first_time, second_time, delay).
        delay: A (possibly negative) integer value representing the number of
               time periods delay in the precedence relationship
    """

    def _to_string(self, values, verbose, smap):
        return self._to_string_impl(values, "==")



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import sys

import pyomo.common.unittest as unittest

from pyomo.environ import (
    ConcreteModel,
    AbstractModel,
    Var,
    Objective,
    Block,
    Constraint,
    Suffix,
    NonNegativeIntegers,
    NonNegativeReals,
    Integers,
    Binary,
    is_fixed,
    value,
)
from pyomo.opt import SolverFactory, TerminationCondition, SolutionStatus
from pyomo.solvers.plugins.solvers.cplex_direct import (
    _CplexExpr,
    _LinearConstraintData,
    _VariableData,
)

try:
    import cplex

    cplexpy_available = True
except ImportError:
    cplexpy_available = False

diff_tol = 1e-4


@unittest.skipIf(not cplexpy_available, "The 'cplex' python bindings are not available")
@unittest.pytest.mark.solver("cplex_direct")
class CPLEXDirectTests(unittest.TestCase):
    def setUp(self):
        self.stderr = sys.stderr
        sys.stderr = None

    def tearDown(self):
        sys.stderr = self.stderr

    def test_infeasible_lp(self):
        with SolverFactory("cplex", solver_io="python") as opt:
            model = ConcreteModel()
            model.X = Var(within=NonNegativeReals)
            model.C1 = Constraint(expr=model.X == 1)
            model.C2 = Constraint(expr=model.X == 2)
            model.O = Objective(expr=model.X)

            results = opt.solve(model)

            self.assertEqual(
                results.solver.termination_condition, TerminationCondition.infeasible
            )

    def test_unbounded_lp(self):
        with SolverFactory("cplex", solver_io="python") as opt:
            model = ConcreteModel()
            model.X = Var()
            model.O = Objective(expr=model.X)

            results = opt.solve(model)

            self.assertIn(
                results.solver.termination_condition,
                (
                    TerminationCondition.unbounded,
                    TerminationCondition.infeasibleOrUnbounded,
                ),
            )

    def test_optimal_lp(self):
        with SolverFactory("cplex", solver_io="python") as opt:
            model = ConcreteModel()
            model.X = Var(within=NonNegativeReals)
            model.O = Objective(expr=model.X)

            results = opt.solve(model, load_solutions=False)

            self.assertEqual(results.solution.status, SolutionStatus.optimal)

    def test_get_duals_lp(self):
        with SolverFactory("cplex", solver_io="python") as opt:
            model = ConcreteModel()
            model.X = Var(within=NonNegativeReals)
            model.Y = Var(within=NonNegativeReals)

            model.C1 = Constraint(expr=2 * model.X + model.Y >= 8)
            model.C2 = Constraint(expr=model.X + 3 * model.Y >= 6)

            model.O = Objective(expr=model.X + model.Y)

            results = opt.solve(model, suffixes=['dual'], load_solutions=False)

            model.dual = Suffix(direction=Suffix.IMPORT)
            model.solutions.load_from(results)

            self.assertAlmostEqual(model.dual[model.C1], 0.4)
            self.assertAlmostEqual(model.dual[model.C2], 0.2)

    def test_infeasible_mip(self):
        with SolverFactory("cplex", solver_io="python") as opt:
            model = ConcreteModel()
            model.X = Var(within=NonNegativeIntegers)
            model.C1 = Constraint(expr=model.X == 1)
            model.C2 = Constraint(expr=model.X == 2)
            model.O = Objective(expr=model.X)

            results = opt.solve(model)

            self.assertEqual(
                results.solver.termination_condition, TerminationCondition.infeasible
            )

    def test_unbounded_mip(self):
        with SolverFactory("cplex", solver_io="python") as opt:
            model = AbstractModel()
            model.X = Var(within=Integers)
            model.O = Objective(expr=model.X)

            instance = model.create_instance()
            results = opt.solve(instance)

            self.assertIn(
                results.solver.termination_condition,
                (
                    TerminationCondition.unbounded,
                    TerminationCondition.infeasibleOrUnbounded,
                ),
            )

    def test_optimal_mip(self):
        with SolverFactory("cplex", solver_io="python") as opt:
            model = ConcreteModel()
            model.X = Var(within=NonNegativeIntegers)
            model.O = Objective(expr=model.X)

            results = opt.solve(model, load_solutions=False)

            self.assertEqual(results.solution.status, SolutionStatus.optimal)


@unittest.skipIf(not cplexpy_available, "The 'cplex' python bindings are not available")
@unittest.pytest.mark.solver("cplex_persistent")
class TestIsFixedCallCount(unittest.TestCase):
    """Tests for PR#1402 (669e7b2b)"""

    def setup(self, skip_trivial_constraints):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.c1 = Constraint(expr=m.x + m.y == 1)
        m.c2 = Constraint(expr=m.x <= 1)
        self.assertFalse(m.c2.has_lb())
        self.assertTrue(m.c2.has_ub())
        self._model = m

        self._opt = SolverFactory("cplex_persistent")
        self._opt.set_instance(
            self._model, skip_trivial_constraints=skip_trivial_constraints
        )

    def test_skip_trivial_and_call_count_for_fixed_con_is_one(self):
        self.setup(skip_trivial_constraints=True)
        self._model.x.fix(1)
        self.assertTrue(self._opt._skip_trivial_constraints)
        self.assertTrue(self._model.c2.body.is_fixed())

        with unittest.mock.patch(
            "pyomo.solvers.plugins.solvers.cplex_direct.is_fixed", wraps=is_fixed
        ) as mock_is_fixed:
            self.assertEqual(mock_is_fixed.call_count, 0)
            self._opt.add_constraint(self._model.c2)
            self.assertEqual(mock_is_fixed.call_count, 1)

    def test_skip_trivial_and_call_count_for_unfixed_con_is_two(self):
        self.setup(skip_trivial_constraints=True)
        self.assertTrue(self._opt._skip_trivial_constraints)
        self.assertFalse(self._model.c2.body.is_fixed())

        with unittest.mock.patch(
            "pyomo.solvers.plugins.solvers.cplex_direct.is_fixed", wraps=is_fixed
        ) as mock_is_fixed:
            self.assertEqual(mock_is_fixed.call_count, 0)
            self._opt.add_constraint(self._model.c2)
            self.assertEqual(mock_is_fixed.call_count, 2)

    def test_skip_trivial_and_call_count_for_unfixed_equality_con_is_three(self):
        self.setup(skip_trivial_constraints=True)
        self._model.c2 = Constraint(expr=self._model.x == 1)
        self.assertTrue(self._opt._skip_trivial_constraints)
        self.assertFalse(self._model.c2.body.is_fixed())

        with unittest.mock.patch(
            "pyomo.solvers.plugins.solvers.cplex_direct.is_fixed", wraps=is_fixed
        ) as mock_is_fixed:
            self.assertEqual(mock_is_fixed.call_count, 0)
            self._opt.add_constraint(self._model.c2)
            self.assertEqual(mock_is_fixed.call_count, 3)

    def test_dont_skip_trivial_and_call_count_for_fixed_con_is_one(self):
        self.setup(skip_trivial_constraints=False)
        self._model.x.fix(1)
        self.assertFalse(self._opt._skip_trivial_constraints)
        self.assertTrue(self._model.c2.body.is_fixed())

        with unittest.mock.patch(
            "pyomo.solvers.plugins.solvers.cplex_direct.is_fixed", wraps=is_fixed
        ) as mock_is_fixed:
            self.assertEqual(mock_is_fixed.call_count, 0)
            self._opt.add_constraint(self._model.c2)
            self.assertEqual(mock_is_fixed.call_count, 1)

    def test_dont_skip_trivial_and_call_count_for_unfixed_con_is_one(self):
        self.setup(skip_trivial_constraints=False)
        self.assertFalse(self._opt._skip_trivial_constraints)
        self.assertFalse(self._model.c2.body.is_fixed())

        with unittest.mock.patch(
            "pyomo.solvers.plugins.solvers.cplex_direct.is_fixed", wraps=is_fixed
        ) as mock_is_fixed:
            self.assertEqual(mock_is_fixed.call_count, 0)
            self._opt.add_constraint(self._model.c2)
            self.assertEqual(mock_is_fixed.call_count, 1)


@unittest.skipIf(not cplexpy_available, "The 'cplex' python bindings are not available")
class TestDataContainers(unittest.TestCase):
    def test_variable_data(self):
        solver_model = cplex.Cplex()
        var_data = _VariableData(solver_model)
        var_data.add(lb=0, ub=1, type_=solver_model.variables.type.binary, name="var1")
        var_data.add(
            lb=0, ub=10, type_=solver_model.variables.type.integer, name="var2"
        )
        var_data.add(
            lb=-cplex.infinity,
            ub=cplex.infinity,
            type_=solver_model.variables.type.continuous,
            name="var3",
        )
        self.assertEqual(solver_model.variables.get_num(), 0)
        var_data.store_in_cplex()
        self.assertEqual(solver_model.variables.get_num(), 3)

    def test_constraint_data(self):
        solver_model = cplex.Cplex()

        solver_model.variables.add(
            lb=[-cplex.infinity, -cplex.infinity, -cplex.infinity],
            ub=[cplex.infinity, cplex.infinity, cplex.infinity],
            types=[
                solver_model.variables.type.continuous,
                solver_model.variables.type.continuous,
                solver_model.variables.type.continuous,
            ],
            names=["var1", "var2", "var3"],
        )
        con_data = _LinearConstraintData(solver_model)
        con_data.add(
            cplex_expr=_CplexExpr(variables=[0, 1], coefficients=[10, 100]),
            sense="L",
            rhs=0,
            range_values=0,
            name="c1",
        )
        con_data.add(
            cplex_expr=_CplexExpr(variables=[0], coefficients=[-30]),
            sense="G",
            rhs=1,
            range_values=0,
            name="c2",
        )
        con_data.add(
            cplex_expr=_CplexExpr(variables=[1], coefficients=[80]),
            sense="E",
            rhs=2,
            range_values=0,
            name="c3",
        )
        con_data.add(
            cplex_expr=_CplexExpr(variables=[2], coefficients=[50]),
            sense="R",
            rhs=3,
            range_values=10,
            name="c4",
        )

        self.assertEqual(solver_model.linear_constraints.get_num(), 0)
        con_data.store_in_cplex()
        self.assertEqual(solver_model.linear_constraints.get_num(), 4)


@unittest.skipIf(not cplexpy_available, "The 'cplex' python bindings are not available")
class TestAddVar(unittest.TestCase):
    def test_add_single_variable(self):
        """Test that the variable is added correctly to `solver_model`."""
        model = ConcreteModel()

        opt = SolverFactory("cplex", solver_io="python")
        opt._set_instance(model)

        self.assertEqual(opt._solver_model.variables.get_num(), 0)
        self.assertEqual(opt._solver_model.variables.get_num_binary(), 0)

        model.X = Var(within=Binary)

        var_interface = opt._solver_model.variables
        with (
            unittest.mock.patch.object(
                var_interface, "add", wraps=var_interface.add
            ) as wrapped_add_call,
            unittest.mock.patch.object(
                var_interface, "set_lower_bounds", wraps=var_interface.set_lower_bounds
            ) as wrapped_lb_call,
            unittest.mock.patch.object(
                var_interface, "set_upper_bounds", wraps=var_interface.set_upper_bounds
            ) as wrapped_ub_call,
        ):
            opt._add_var(model.X)

            self.assertEqual(wrapped_add_call.call_count, 1)
            self.assertEqual(
                wrapped_add_call.call_args,
                ({"lb": [0], "names": ["x1"], "types": ["B"], "ub": [1]},),
            )

            self.assertFalse(wrapped_lb_call.called)
            self.assertFalse(wrapped_ub_call.called)

        self.assertEqual(opt._solver_model.variables.get_num(), 1)
        self.assertEqual(opt._solver_model.variables.get_num_binary(), 1)

    def test_add_block_containing_single_variable(self):
        """Test that the variable is added correctly to `solver_model`."""
        model = ConcreteModel()

        opt = SolverFactory("cplex", solver_io="python")
        opt._set_instance(model)

        self.assertEqual(opt._solver_model.variables.get_num(), 0)
        self.assertEqual(opt._solver_model.variables.get_num_binary(), 0)

        model.X = Var(within=Binary)

        with unittest.mock.patch.object(
            opt._solver_model.variables, "add", wraps=opt._solver_model.variables.add
        ) as wrapped_add_call:
            opt._add_block(model)

            self.assertEqual(wrapped_add_call.call_count, 1)
            self.assertEqual(
                wrapped_add_call.call_args,
                ({"lb": [0], "names": ["x1"], "types": ["B"], "ub": [1]},),
            )

        self.assertEqual(opt._solver_model.variables.get_num(), 1)
        self.assertEqual(opt._solver_model.variables.get_num_binary(), 1)

    def test_add_block_containing_multiple_variables(self):
        """Test that:
        - The variable is added correctly to `solver_model`
        - The CPLEX `variables` interface is called only once
        - Fixed variable bounds are set correctly
        """
        model = ConcreteModel()

        opt = SolverFactory("cplex", solver_io="python")
        opt._set_instance(model)

        self.assertEqual(opt._solver_model.variables.get_num(), 0)

        model.X1 = Var(within=Binary)
        model.X2 = Var(within=NonNegativeReals)
        model.X3 = Var(within=NonNegativeIntegers)

        model.X3.fix(5)

        with unittest.mock.patch.object(
            opt._solver_model.variables, "add", wraps=opt._solver_model.variables.add
        ) as wrapped_add_call:
            opt._add_block(model)

            self.assertEqual(wrapped_add_call.call_count, 1)
            self.assertEqual(
                wrapped_add_call.call_args,
                (
                    {
                        "lb": [0, 0, 5],
                        "names": ["x1", "x2", "x3"],
                        "types": ["B", "C", "I"],
                        "ub": [1, cplex.infinity, 5],
                    },
                ),
            )

        self.assertEqual(opt._solver_model.variables.get_num(), 3)


@unittest.skipIf(not cplexpy_available, "The 'cplex' python bindings are not available")
@unittest.pytest.mark.solver("cplex_direct")
class TestAddCon(unittest.TestCase):
    def test_add_single_constraint(self):
        model = ConcreteModel()
        model.X = Var(within=Binary)

        opt = SolverFactory("cplex", solver_io="python")
        opt._set_instance(model)

        self.assertEqual(opt._solver_model.linear_constraints.get_num(), 0)

        model.C = Constraint(expr=model.X == 1)

        con_interface = opt._solver_model.linear_constraints
        with unittest.mock.patch.object(
            con_interface, "add", wraps=con_interface.add
        ) as wrapped_add_call:
            opt._add_constraint(model.C)

            self.assertEqual(wrapped_add_call.call_count, 1)
            self.assertEqual(
                wrapped_add_call.call_args,
                (
                    {
                        "lin_expr": [[[0], (1,)]],
                        "names": ["x2"],
                        "range_values": [0.0],
                        "rhs": [1.0],
                        "senses": ["E"],
                    },
                ),
            )

        self.assertEqual(opt._solver_model.linear_constraints.get_num(), 1)

    def test_add_block_containing_single_constraint(self):
        model = ConcreteModel()
        model.X = Var(within=Binary)

        opt = SolverFactory("cplex", solver_io="python")
        opt._set_instance(model)

        self.assertEqual(opt._solver_model.linear_constraints.get_num(), 0)

        model.B = Block()
        model.B.C = Constraint(expr=model.X == 1)

        con_interface = opt._solver_model.linear_constraints
        with unittest.mock.patch.object(
            con_interface, "add", wraps=con_interface.add
        ) as wrapped_add_call:
            opt._add_block(model.B)

            self.assertEqual(wrapped_add_call.call_count, 1)
            self.assertEqual(
                wrapped_add_call.call_args,
                (
                    {
                        "lin_expr": [[[0], (1,)]],
                        "names": ["x2"],
                        "range_values": [0.0],
                        "rhs": [1.0],
                        "senses": ["E"],
                    },
                ),
            )

        self.assertEqual(opt._solver_model.linear_constraints.get_num(), 1)

    def test_add_block_containing_multiple_constraints(self):
        model = ConcreteModel()
        model.X = Var(within=Binary)

        opt = SolverFactory("cplex", solver_io="python")
        opt._set_instance(model)

        self.assertEqual(opt._solver_model.linear_constraints.get_num(), 0)

        model.B = Block()
        model.B.C1 = Constraint(expr=model.X == 1)
        model.B.C2 = Constraint(expr=model.X <= 1)
        model.B.C3 = Constraint(expr=model.X >= 1)

        con_interface = opt._solver_model.linear_constraints
        with unittest.mock.patch.object(
            con_interface, "add", wraps=con_interface.add
        ) as wrapped_add_call:
            opt._add_block(model.B)

            self.assertEqual(wrapped_add_call.call_count, 1)
            self.assertEqual(
                wrapped_add_call.call_args,
                (
                    {
                        "lin_expr": [[[0], (1,)], [[0], (1,)], [[0], (1,)]],
                        "names": ["x2", "x3", "x4"],
                        "range_values": [0.0, 0.0, 0.0],
                        "rhs": [1.0, 1.0, 1.0],
                        "senses": ["E", "L", "G"],
                    },
                ),
            )

        self.assertEqual(opt._solver_model.linear_constraints.get_num(), 3)


@unittest.skipIf(not cplexpy_available, "The 'cplex' python bindings are not available")
@unittest.pytest.mark.solver("cplex_direct")
class TestLoadVars(unittest.TestCase):
    def setUp(self):
        opt = SolverFactory("cplex", solver_io="python")
        model = ConcreteModel()
        model.X = Var(within=NonNegativeReals, initialize=0)
        model.Y = Var(within=NonNegativeReals, initialize=0)

        model.C1 = Constraint(expr=2 * model.X + model.Y >= 8)
        model.C2 = Constraint(expr=model.X + 3 * model.Y >= 6)

        model.O = Objective(expr=model.X + model.Y)

        opt.solve(model, load_solutions=False, save_results=False)

        self._model = model
        self._opt = opt

    def test_all_vars_are_loaded(self):
        self.assertTrue(self._model.X.stale)
        self.assertTrue(self._model.Y.stale)
        self.assertEqual(value(self._model.X), 0)
        self.assertEqual(value(self._model.Y), 0)

        with unittest.mock.patch.object(
            self._opt._solver_model.solution,
            "get_values",
            wraps=self._opt._solver_model.solution.get_values,
        ) as wrapped_values_call:
            self._opt.load_vars()

            self.assertEqual(wrapped_values_call.call_count, 1)
            self.assertEqual(wrapped_values_call.call_args, tuple())

        self.assertFalse(self._model.X.stale)
        self.assertFalse(self._model.Y.stale)
        self.assertAlmostEqual(value(self._model.X), 3.6)
        self.assertAlmostEqual(value(self._model.Y), 0.8)

    def test_only_specified_vars_are_loaded(self):
        self.assertTrue(self._model.X.stale)
        self.assertTrue(self._model.Y.stale)
        self.assertEqual(value(self._model.X), 0)
        self.assertEqual(value(self._model.Y), 0)

        with unittest.mock.patch.object(
            self._opt._solver_model.solution,
            "get_values",
            wraps=self._opt._solver_model.solution.get_values,
        ) as wrapped_values_call:
            self._opt.load_vars([self._model.X])

            self.assertEqual(wrapped_values_call.call_count, 1)
            self.assertEqual(wrapped_values_call.call_args, (([0],), {}))

        self.assertFalse(self._model.X.stale)
        self.assertTrue(self._model.Y.stale)
        self.assertAlmostEqual(value(self._model.X), 3.6)
        self.assertEqual(value(self._model.Y), 0)

        with unittest.mock.patch.object(
            self._opt._solver_model.solution,
            "get_values",
            wraps=self._opt._solver_model.solution.get_values,
        ) as wrapped_values_call:
            self._opt.load_vars([self._model.Y])

            self.assertEqual(wrapped_values_call.call_count, 1)
            self.assertEqual(wrapped_values_call.call_args, (([1],), {}))

        self.assertFalse(self._model.X.stale)
        self.assertFalse(self._model.Y.stale)
        self.assertAlmostEqual(value(self._model.X), 3.6)
        self.assertAlmostEqual(value(self._model.Y), 0.8)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Development of this module was conducted as part of the Institute for
# the Design of Advanced Energy Systems (IDAES) with support through the
# Simulation-Based Engineering, Crosscutting Research Program within the
# U.S. Department of Energy's Office of Fossil Energy and Carbon Management.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.contrib.trustregion.filter import Filter, FilterElement


class TestFilter(unittest.TestCase):
    def setUp(self):
        self.objective = 1.0
        self.feasible = 0.5
        self.theta_max = 10.0
        self.tmpFilter = Filter()

    def tearDown(self):
        pass

    def test_FilterElement(self):
        fe = FilterElement(self.objective, self.feasible)
        self.assertEqual(fe.objective, self.objective)
        self.assertEqual(fe.feasible, self.feasible)

    def test_addToFilter(self):
        fe = FilterElement(self.objective, self.feasible)
        self.tmpFilter.addToFilter(fe)
        self.assertIn(fe, self.tmpFilter.TrustRegionFilter)

    def test_isAcceptable(self):
        fe = FilterElement(0.5, 0.25)
        # A sufficiently feasible element
        self.assertTrue(self.tmpFilter.isAcceptable(fe, self.theta_max))
        fe = FilterElement(10.0, 15.0)
        # A sufficiently infeasible element
        self.assertFalse(self.tmpFilter.isAcceptable(fe, self.theta_max))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# -*- coding: UTF-8 -*-
"""Module with miscellaneous diagnostic tools"""

from pyomo.core.base.block import TraversalStrategy, Block
from pyomo.gdp import Disjunct
import logging

logger = logging.getLogger('pyomo.util.diagnostics')
logger.setLevel(logging.INFO)


def log_disjunct_values(m):
    """Prints the values of the disjunct indicator variables."""
    for disj in m.component_data_objects(
        ctype=Disjunct,
        active=True,
        descend_into=(Block, Disjunct),
        descent_order=TraversalStrategy.PrefixDepthFirstSearch,
    ):
        logger.info(
            "%s %s%s"
            % (
                disj.name,
                disj.indicator_var.value,
                " fixed" if disj.indicator_var.fixed else "",
            )
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import inspect
import textwrap


def format_exception(msg, prolog=None, epilog=None, exception=None, width=76):
    """Generate a formatted exception message

    This returns a formatted exception message, line wrapped for display
    on the console and with optional prolog and epilog messages.

    Parameters
    ----------
    msg: str
        The raw exception message

    prolog: str, optional
        A message to output before the exception message, ``msg``.  If
        this message is long enough to line wrap, the ``msg`` will be
        indented a level below the ``prolog`` message.

    epilog: str, optional
        A message to output after the exception message, ``msg``.  If
        provided, the ``msg`` will be indented a level below the
        ``prolog`` / ``epilog`` messages.

    exception: Exception, optional
        The raw exception being raised (used to improve initial line wrapping).

    width: int, optional
        The line length to wrap the exception message to.

    Returns
    -------
    str
    """
    fields = []

    if epilog:
        indent = ' ' * 8
    else:
        indent = ' ' * 4

    if exception is None:
        # default to the length of 'NotImplementedError: ', the longest
        # built-in name that we commonly raise
        initial_indent = ' ' * 21
    else:
        if not inspect.isclass(exception):
            exception = exception.__class__
        initial_indent = ' ' * (len(exception.__name__) + 2)
        if exception.__module__ != 'builtins':
            initial_indent += ' ' * (len(exception.__module__) + 1)

    if prolog is not None:
        if '\n' not in prolog:
            # We want to strip off the leading indent that we added as a
            # placeholder for the string representation of the exception
            # class name.
            prolog = textwrap.fill(
                prolog,
                width=width,
                initial_indent=initial_indent,
                subsequent_indent=' ' * 4,
                break_long_words=False,
                break_on_hyphens=False,
            ).lstrip()
        # If the prolog line-wrapped, ensure that the message is
        # indented an additional level.
        if '\n' in prolog:
            indent = ' ' * 8
        fields.append(prolog)
        initial_indent = indent

    if '\n' not in msg:
        msg = textwrap.fill(
            msg,
            width=width,
            initial_indent=initial_indent,
            subsequent_indent=indent,
            break_long_words=False,
            break_on_hyphens=False,
        )
        if not fields:
            # We want to strip off the leading indent that we just
            # added, but only if there is no prolog
            msg = msg.lstrip()
    if msg:
        fields.append(msg)

    if epilog is not None:
        if '\n' not in epilog:
            epilog = textwrap.fill(
                epilog,
                width=width,
                initial_indent=' ' * 4,
                subsequent_indent=' ' * 4,
                break_long_words=False,
                break_on_hyphens=False,
            )
        fields.append(epilog)

    return '\n'.join(fields)


class ApplicationError(Exception):
    """
    An exception used when an external application generates an error.
    """


class PyomoException(Exception):
    """
    Exception class for other Pyomo exceptions to inherit from,
    allowing Pyomo exceptions to be caught in a general way
    (e.g., in other applications that use Pyomo).
    Subclasses can define a class-level `default_message` attribute.
    """

    def __init__(self, *args):
        if not args and getattr(self, 'default_message', None):
            args = (self.default_message,)
        return super().__init__(*args)


class DeferredImportError(ImportError):
    """This exception is raised when something attempts to access a module
    that was imported by :py:func:`.attempt_import`, but the module
    import failed.

    """


class DeveloperError(PyomoException, NotImplementedError):
    """
    Exception class used to throw errors that result from Pyomo
    programming errors, rather than user modeling errors (e.g., a
    component not declaring a 'ctype').
    """

    def __str__(self):
        msg = super().__str__()
        if msg:
            msg = repr(msg)
        return format_exception(
            msg,
            prolog="Internal Pyomo implementation error:",
            epilog="Please report this to the Pyomo Developers.",
            exception=self,
        )


class InfeasibleConstraintException(PyomoException):
    """
    Exception class used by Pyomo transformations to indicate
    that an infeasible constraint has been identified (e.g. in
    the course of range reduction).
    """


class IterationLimitError(PyomoException, RuntimeError):
    """A subclass of :py:class:`RuntimeError`, raised by an iterative method
    when the iteration limit is reached.

    TODO: solvers currently do not raise this exception, but probably
    should (at least when non-normal termination conditions are mapped
    to exceptions)

    """


class IntervalException(PyomoException, ValueError):
    """
    Exception class used for errors in interval arithmetic.
    """


class InvalidValueError(PyomoException, ValueError):
    """
    Exception class used for value errors in compiled model representations
    """


class MouseTrap(PyomoException, NotImplementedError):
    """
    Exception class used to throw errors for not-implemented functionality
    that might be rational to support (i.e., we already gave you a cookie)
    but risks taking Pyomo's flexibility a step beyond what is sane,
    or solvable, or communicable to a solver, etc. (i.e., Really? Now you
    want a glass of milk too?)
    """

    def __str__(self):
        return format_exception(
            repr(super().__str__()),
            prolog="Sorry, mouse, no cookies here!",
            epilog="This is functionality we think may be rational to "
            "support, but is not yet implemented (possibly due to developer "
            "availability, complexity of edge cases, or general practicality "
            "or tractability). However, please feed the mice: "
            "pull requests are always welcome!",
            exception=self,
        )


class NondifferentiableError(PyomoException, ValueError):
    """A Pyomo-specific ValueError raised for non-differentiable expressions"""


class TempfileContextError(PyomoException, IndexError):
    """A Pyomo-specific IndexError raised when attempting to use the
    TempfileManager when it does not have a currently active context.

    """


class TemplateExpressionError(ValueError):
    """Special ValueError raised by getitem for template arguments

    This exception is triggered by the Pyomo expression system when
    attempting to get a member of an IndexedComponent using either a
    TemplateIndex, or an expression containing a TemplateIndex.

    Users should never see this exception.

    """

    def __init__(self, template, *args, **kwds):
        self.template = template
        super(TemplateExpressionError, self).__init__(*args, **kwds)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
from pyomo.common.dependencies import numpy as np, pathlib

from pyomo.contrib.doe.examples.reactor_experiment import ReactorExperiment
from pyomo.contrib.doe import DesignOfExperiments

import pyomo.environ as pyo

import json


# Example for sensitivity analysis on the reactor experiment
# After sensitivity analysis is done, we perform optimal DoE
def run_reactor_doe(
    n_points_for_design=9,
    compute_FIM_full_factorial=True,
    plot_factorial_results=True,
    figure_file_name="example_reactor_compute_FIM",
    log_scale=False,
    run_optimal_doe=True,
):
    """
    This function demonstrates how to perform sensitivity analysis on the reactor

    Parameters
    ----------
    n_points_for_design : int, optional
        number of points to use for the design ranges, by default 9
    compute_FIM_full_factorial : bool, optional
        whether to compute the full factorial design, by default True
    plot_factorial_results : bool, optional
        whether to plot the results of the full factorial design, by default True
    figure_file_name : str, optional
        file name to save the factorial figure, by default "example_reactor_compute_FIM"
    run_optimal_doe : bool, optional
        whether to run the optimal DoE, by default True
    """
    # Read in file
    DATA_DIR = pathlib.Path(__file__).parent
    file_path = DATA_DIR / "result.json"

    with open(file_path) as f:
        data_ex = json.load(f)

    # Put temperature control time points into correct format for reactor experiment
    data_ex["control_points"] = {
        float(k): v for k, v in data_ex["control_points"].items()
    }

    # Create a ReactorExperiment object; data and discretization information are part
    # of the constructor of this object
    experiment = ReactorExperiment(data=data_ex, nfe=10, ncp=3)

    # Use a central difference, with step size 1e-3
    fd_formula = "central"
    step_size = 1e-3

    # Use the determinant objective with scaled sensitivity matrix
    objective_option = "determinant"
    scale_nominal_param_value = True

    # Create the DesignOfExperiments object
    # We will not be passing any prior information in this example
    # and allow the experiment object and the DesignOfExperiments
    # call of ``run_doe`` perform model initialization.
    doe_obj = DesignOfExperiments(
        experiment,
        fd_formula=fd_formula,
        step=step_size,
        objective_option=objective_option,
        scale_constant_value=1,
        scale_nominal_param_value=scale_nominal_param_value,
        prior_FIM=None,
        jac_initial=None,
        fim_initial=None,
        L_diagonal_lower_bound=1e-7,
        solver=None,
        tee=False,
        get_labeled_model_args=None,
        _Cholesky_option=True,
        _only_compute_fim_lower=True,
    )
    if compute_FIM_full_factorial:
        # Make design ranges to compute the full factorial design
        design_ranges = {
            "CA[0]": [1, 5, n_points_for_design],
            "T[0]": [300, 700, n_points_for_design],
        }

        # Compute the full factorial design with the sequential FIM calculation
        doe_obj.compute_FIM_full_factorial(
            design_ranges=design_ranges, method="sequential"
        )
    if plot_factorial_results:
        # Plot the results
        doe_obj.draw_factorial_figure(
            sensitivity_design_variables=["CA[0]", "T[0]"],
            fixed_design_variables={
                "T[0.125]": 300,
                "T[0.25]": 300,
                "T[0.375]": 300,
                "T[0.5]": 300,
                "T[0.625]": 300,
                "T[0.75]": 300,
                "T[0.875]": 300,
                "T[1]": 300,
            },
            title_text="Reactor Example",
            xlabel_text="Concentration of A (M)",
            ylabel_text="Initial Temperature (K)",
            figure_file_name=figure_file_name,
            log_scale=log_scale,
        )

    ###########################
    # End sensitivity analysis

    # Begin optimal DoE
    ####################
    if run_optimal_doe:
        doe_obj.run_doe()

        # Print out a results summary
        print("Optimal experiment values: ")
        print(
            "\tInitial concentration: {:.2f}".format(
                doe_obj.results["Experiment Design"][0]
            )
        )
        print(
            ("\tTemperature values: [" + "{:.2f}, " * 8 + "{:.2f}]").format(
                *doe_obj.results["Experiment Design"][1:]
            )
        )
        print("FIM at optimal design:\n {}".format(np.array(doe_obj.results["FIM"])))
        print(
            "Objective value at optimal design: {:.2f}".format(
                pyo.value(doe_obj.model.objective)
            )
        )

        print(doe_obj.results["Experiment Design Names"])

        ###################
        # End optimal DoE

    return doe_obj


if __name__ == "__main__":
    run_reactor_doe()
